# future
from __future__ import annotations

# stdlib
from enum import Enum
from enum import auto
from typing import Any

# relative
from ...serde.serializable import serializable
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.uid import UID
from ..context import AuthedServiceContext


class EXECUTION_MODE(Enum):
    CALL = auto()
    MOCK = auto()
    PRIVATE = auto()


@serializable()
class CustomEndpointActionObject(SyftObject):
    __canonical_name__ = "CustomEndpointActionObject"
    __version__ = SYFT_OBJECT_VERSION_1

    endpoint_id: UID
    context: AuthedServiceContext | None = None

    def add_context(self, context: AuthedServiceContext) -> CustomEndpointActionObject:
        self.context = context
        return self

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.__call_function(  #  type: ignore[misc]
            *args,
            **kwargs,
            call_mode=EXECUTION_MODE.CALL,
        )

    def mock(self, *args: Any, **kwargs: Any) -> Any:
        return self.__call_function(  #  type: ignore[misc]
            *args,
            **kwargs,
            call_mode=EXECUTION_MODE.MOCK,
        )

    def private(self, *args: Any, **kwargs: Any) -> Any:
        return self.__call_function(  #  type: ignore[misc]
            *args,
            **kwargs,
            call_mode=EXECUTION_MODE.PRIVATE,
        )

    def __call_function(
        self, call_mode: EXECUTION_MODE, *args: Any, **kwargs: Any
    ) -> Any:
        self.context = self.__check_context()
        endpoint_service = self.context.server.get_service("apiservice")

        if call_mode == EXECUTION_MODE.MOCK:
            __endpoint_mode = endpoint_service.execute_server_side_endpoint_mock_by_id
        elif call_mode == EXECUTION_MODE.PRIVATE:
            __endpoint_mode = (
                endpoint_service.execute_service_side_endpoint_private_by_id
            )
        else:
            __endpoint_mode = endpoint_service.execute_server_side_endpoint_by_id

        return __endpoint_mode(
            *args, context=self.context, endpoint_uid=self.endpoint_id, **kwargs
        )

    def __check_context(self) -> AuthedServiceContext:
        if self.context is None:
            raise Exception("No context provided to CustomEndpointActionObject")
        return self.context



# stdlib
import contextlib
from hashlib import sha256
import io
from pathlib import Path
from typing import Any

# third party
import docker
from packaging import version
from pydantic import field_validator
from typing_extensions import Self
import yaml

# relative
from ..serde.serializable import serializable
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..types.base import SyftBaseModel
from .utils import iterator_to_string

PYTHON_DEFAULT_VER = "3.12"
PYTHON_MIN_VER = version.parse("3.10")
PYTHON_MAX_VER = version.parse("3.12")


def _malformed_python_package_error_msg(pkg: str, name: str = "package_name") -> str:
    return f'You must pin the package to an exact version. Got "{pkg}" expected "{name}==x.y.z"'


class CustomBuildConfig(SyftBaseModel):
    gpu: bool = False
    # python_version: str = PYTHON_DEFAULT_VER
    python_packages: list[str] = []
    system_packages: list[str] = []
    custom_cmds: list[str] = []

    # @validator("python_version")
    # def validate_python_version(cls, ver: str) -> str:
    #     parsed_ver = version.parse(ver)

    #     # TODO: Check if Wolfi OS/apk supports minor version of python
    #     if parsed_ver.micro != 0:
    #         raise ValueError("Provide only major.minor version of python")

    #     if PYTHON_MIN_VER <= parsed_ver < PYTHON_MAX_VER:
    #         return ver
    #     else:
    #         raise ValueError(
    #             f"Python version must be between {PYTHON_MIN_VER} and {PYTHON_MAX_VER}"
    #         )

    @field_validator("python_packages")
    @classmethod
    def validate_python_packages(cls, pkgs: list[str]) -> list[str]:
        for pkg in pkgs:
            ver_parts: tuple | list = ()
            name_ver = pkg.split("==")
            if len(name_ver) != 2:
                raise ValueError(_malformed_python_package_error_msg(pkg))

            name, ver = name_ver

            if ver:
                ver_parts = ver.split(".")

            if not ver or len(ver_parts) <= 2:
                raise ValueError(_malformed_python_package_error_msg(pkg, name))

        return pkgs

    def merged_python_pkgs(self, sep: str = " ") -> str:
        return sep.join(self.python_packages)

    def merged_system_pkgs(self, sep: str = " ") -> str:
        return sep.join(self.system_packages)

    def merged_custom_cmds(self, sep: str = ";") -> str:
        return sep.join(self.custom_cmds)


class WorkerConfig(SyftBaseModel):
    pass


@serializable(canonical_name="CustomWorkerConfig", version=1)
class CustomWorkerConfig(WorkerConfig):
    build: CustomBuildConfig
    version: str = "1"

    @classmethod
    def from_dict(cls, config: dict[str, Any]) -> Self:
        return cls(**config)

    @classmethod
    def from_str(cls, content: str) -> Self:
        config = yaml.safe_load(content)
        return cls.from_dict(config)

    @classmethod
    def from_path(cls, path: Path | str) -> Self:
        with open(path) as f:
            config = yaml.safe_load(f)
            return cls.from_dict(config)

    def get_signature(self) -> str:
        return sha256(self.json(sort_keys=True).encode()).hexdigest()


@serializable(canonical_name="PrebuiltWorkerConfig", version=1)
class PrebuiltWorkerConfig(WorkerConfig):
    # tag that is already built and pushed in some registry
    tag: str
    description: str | None = None

    def __str__(self) -> str:
        if self.description:
            return f"prebuilt tag='{self.tag}' description='{self.description}'"
        else:
            return f"prebuilt tag='{self.tag}'"

    def set_description(self, description_text: str) -> None:
        self.description = description_text

    def __hash__(self) -> int:
        return hash(self.tag)


@serializable(canonical_name="DockerWorkerConfig", version=1)
class DockerWorkerConfig(WorkerConfig):
    dockerfile: str
    file_name: str | None = None
    description: str | None = None

    @field_validator("dockerfile")
    @classmethod
    def validate_dockerfile(cls, dockerfile: str) -> str:
        if not dockerfile:
            raise ValueError("Dockerfile cannot be empty")
        dockerfile = dockerfile.strip()
        return dockerfile

    @classmethod
    def from_path(
        cls,
        path: Path | str,
        description: str | None = "",
    ) -> Self:
        with open(path) as f:
            return cls(
                dockerfile=f.read(),
                file_name=Path(path).name,
                description=description,
            )

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, DockerWorkerConfig):
            return False
        return self.dockerfile == __value.dockerfile

    def __hash__(self) -> int:
        return hash(self.dockerfile)

    def __str__(self) -> str:
        return self.dockerfile

    def set_description(self, description_text: str) -> None:
        self.description = description_text

    def test_image_build(self, tag: str, **kwargs: Any) -> SyftSuccess | SyftError:
        try:
            with contextlib.closing(docker.from_env()) as client:
                if not client.ping():
                    return SyftError(
                        "Cannot reach docker server. Please check if docker is running."
                    )

                kwargs["fileobj"] = io.BytesIO(self.dockerfile.encode("utf-8"))
                _, logs = client.images.build(
                    tag=tag,
                    rm=True,
                    labels={"orgs.openmined.syft": "Test image build"},
                    **kwargs,
                )
                return SyftSuccess(message=iterator_to_string(iterator=logs))
        except Exception as e:
            # stdlib
            import traceback

            return SyftError(message=f"Failed to build: {e} {traceback.format_exc()}")



# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_permissions import StoragePermission
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from .log import SyftLog
from .log_stash import LogStash


@instrument
@serializable(canonical_name="LogService", version=1)
class LogService(AbstractService):
    store: DocumentStore
    stash: LogStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = LogStash(store=store)

    @service_method(path="log.add", name="add", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def add(
        self,
        context: AuthedServiceContext,
        uid: UID,
        job_id: UID,
        stdout: str = "",
        stderr: str = "",
    ) -> SyftSuccess | SyftError:
        new_log = SyftLog(id=uid, job_id=job_id, stdout=stdout, stderr=stderr)
        result = self.stash.set(context.credentials, new_log)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result

    @service_method(path="log.append", name="append", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def append(
        self,
        context: AuthedServiceContext,
        uid: UID,
        new_str: str = "",
        new_err: str = "",
    ) -> SyftSuccess | SyftError:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))
        new_log = result.ok()
        if new_str:
            new_log.append(new_str)

        if new_err:
            new_log.append_error(new_err)

        result = self.stash.update(context.credentials, new_log)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="Log Append successful!")

    @service_method(path="log.get", name="get", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def get(self, context: AuthedServiceContext, uid: UID) -> SyftLog | SyftError:
        result = self.stash.get_by_uid(context.credentials, uid)

        if result.is_err():
            return SyftError(message=str(result.err()))

        return result.ok()

    @service_method(
        path="log.get_stdout", name="get_stdout", roles=DATA_SCIENTIST_ROLE_LEVEL
    )
    def get_stdout(self, context: AuthedServiceContext, uid: UID) -> str | SyftError:
        result = self.get(context, uid)
        if isinstance(result, SyftError):
            return result
        return result.stdout

    @service_method(path="log.get_stderr", name="get_stderr", roles=ADMIN_ROLE_LEVEL)
    def get_stderr(self, context: AuthedServiceContext, uid: UID) -> str | SyftError:
        result = self.get(context, uid)
        if isinstance(result, SyftError):
            return result
        return result.stderr

    @service_method(path="log.restart", name="restart", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def restart(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> SyftSuccess | SyftError:
        result = self.stash.get_by_uid(context.credentials, uid)
        if result.is_err():
            return SyftError(message=str(result.err()))

        log = result.ok()
        log.restart()
        result = self.stash.update(context.credentials, log)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return SyftSuccess(message="Log Restart successful!")

    @service_method(path="log.get_all", name="get_all", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def get_all(self, context: AuthedServiceContext) -> SyftSuccess | SyftError:
        result = self.stash.get_all(context.credentials)
        if result.is_err():
            return SyftError(message=str(result.err()))
        return result.ok()

    @service_method(path="log.delete", name="delete", roles=DATA_SCIENTIST_ROLE_LEVEL)
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        result = self.stash.delete_by_uid(context.credentials, uid)
        if result.is_ok():
            return result.ok()
        else:
            return SyftError(message=result.err())

    @service_method(
        path="log.has_storage_permission",
        name="has_storage_permission",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def has_storage_permission(self, context: AuthedServiceContext, uid: UID) -> bool:
        permission = StoragePermission(uid=uid, server_uid=context.server.id)
        result = self.stash.has_storage_permission(permission)

        return result


TYPE_TO_SERVICE[SyftLog] = LogService



# stdlib

# relative
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ..service import AbstractService


@serializable(canonical_name="EnclaveService", version=1)
class EnclaveService(AbstractService):
    store: DocumentStore

    def __init__(self, store: DocumentStore) -> None:
        self.store = store



# stdlib
from collections import defaultdict
import logging
import threading
import time
from typing import Any

# third party
from pydantic import BaseModel
from sherlock.lock import BaseLock

# relative
from ..serde.serializable import serializable

logger = logging.getLogger(__name__)
THREAD_FILE_LOCKS: dict[int, dict[str, int]] = defaultdict(dict)


@serializable(canonical_name="LockingConfig", version=1)
class LockingConfig(BaseModel):
    """
    Locking config

    Args:
        lock_name: str
            Lock name
        namespace: Optional[str]
            Namespace to use for setting lock keys in the backend store.
        expire: Optional[int]
            Lock expiration time in seconds. If explicitly set to `None`, lock will not expire.
        timeout: Optional[int]
             Timeout to acquire lock(seconds)
        retry_interval: float
            Retry interval to retry acquiring a lock if previous attempts failed.
    """

    lock_name: str = "syft_lock"
    namespace: str | None = None
    expire: int | None = 60
    timeout: int | None = 30
    retry_interval: float = 0.1


@serializable(canonical_name="NoLockingConfig", version=1)
class NoLockingConfig(LockingConfig):
    """
    No-locking policy
    """

    pass


@serializable(canonical_name="ThreadingLockingConfig", version=1)
class ThreadingLockingConfig(LockingConfig):
    """
    Threading-based locking policy
    """

    pass


class ThreadingLock(BaseLock):
    """
    Threading-based Lock. Used to provide the same API as the rest of the locks.
    """

    def __init__(self, expire: int, **kwargs: Any) -> None:
        self.expire = expire
        self.locked_timestamp: float = 0.0
        self.lock = threading.Lock()

    @property
    def _locked(self) -> bool:
        """
        Implementation of method to check if lock has been acquired. Must be
        :returns: if the lock is acquired or not
        :rtype: bool
        """
        locked = self.lock.locked()
        if (
            locked
            and time.time() - self.locked_timestamp >= self.expire
            and self.expire != -1
        ):
            self._release()

        return self.lock.locked()

    def _acquire(self) -> bool:
        """
        Implementation of acquiring a lock in a non-blocking fashion.
        :returns: if the lock was successfully acquired or not
        :rtype: bool
        """
        locked = self.lock.locked()
        if (
            locked
            and time.time() - self.locked_timestamp > self.expire
            and self.expire != -1
        ):
            self._release()

        status = self.lock.acquire(
            blocking=False
        )  # timeout/retries handle in the `acquire` method
        if status:
            self.locked_timestamp = time.time()
        return status

    def _release(self) -> None:
        """
        Implementation of releasing an acquired lock.
        """

        try:
            return self.lock.release()
        except RuntimeError:  # already unlocked
            pass

    def _renew(self) -> bool:
        """
        Implementation of renewing an acquired lock.
        """
        return True


class SyftLock(BaseLock):
    """
    Syft Lock implementations.

    Params:
        config: Config specific to a locking strategy.
    """

    def __init__(self, config: LockingConfig):
        self.config = config

        self.lock_name = config.lock_name
        self.namespace = config.namespace
        self.expire = config.expire
        self.timeout = config.timeout
        self.retry_interval = config.retry_interval

        self.passthrough = False

        self._lock: BaseLock | None = None

        base_params = {
            "lock_name": config.lock_name,
            "namespace": config.namespace,
            "expire": config.expire,
            "timeout": config.timeout,
            "retry_interval": config.retry_interval,
        }
        if isinstance(config, NoLockingConfig):
            self.passthrough = True
        elif isinstance(config, ThreadingLockingConfig):
            self._lock = ThreadingLock(**base_params)
        else:
            raise ValueError("Unsupported config type")

    @property
    def _locked(self) -> bool:
        """
        Implementation of method to check if lock has been acquired.

        :returns: if the lock is acquired or not
        :rtype: bool
        """
        if self.passthrough:
            return False
        return self._lock.locked() if self._lock else False

    def acquire(self, blocking: bool = True) -> bool:
        """
        Acquire a lock, blocking or non-blocking.
        :param bool blocking: acquire a lock in a blocking or non-blocking
                              fashion. Defaults to True.
        :returns: if the lock was successfully acquired or not
        :rtype: bool
        """

        if not blocking:
            return self._acquire()

        timeout: float = float(self.timeout)
        start_time = time.time()
        elapsed: float = 0.0
        while timeout >= elapsed:
            if not self._acquire():
                time.sleep(self.retry_interval)
                elapsed = time.time() - start_time
            else:
                return True
        logger.debug(
            f"Timeout elapsed after {self.timeout} seconds while trying to acquiring lock."
        )
        # third party
        return False

    def _acquire(self) -> bool:
        """
        Implementation of acquiring a lock in a non-blocking fashion.
        `acquire` makes use of this implementation to provide blocking and non-blocking implementations.

        :returns: if the lock was successfully acquired or not
        :rtype: bool
        """
        if self.passthrough:
            return True

        try:
            return self._lock._acquire() if self._lock else False
        except BaseException:
            return False

    def _release(self) -> bool | None:
        """
        Implementation of releasing an acquired lock.
        """
        if self.passthrough:
            return None
        if not self._lock:
            return None
        try:
            return self._lock._release()
        except BaseException:
            return None

    def _renew(self) -> bool:
        """
        Implementation of renewing an acquired lock.
        """
        if self.passthrough:
            return True

        return self._lock._renew() if self._lock else False



# future
from __future__ import annotations

# stdlib
from typing import Any

# third party
from pydantic import Field

# relative
from ..serde.serializable import serializable
from ..server.credentials import SyftVerifyKey
from ..types import uid
from .document_store import DocumentStore
from .document_store import StoreConfig
from .kv_document_store import KeyValueBackingStore
from .kv_document_store import KeyValueStorePartition
from .locks import LockingConfig
from .locks import ThreadingLockingConfig


@serializable(canonical_name="DictBackingStore", version=1)
class DictBackingStore(dict, KeyValueBackingStore):  # type: ignore[misc]
    # TODO: fix the mypy issue
    """Dictionary-based Store core logic"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__()
        self._ddtype = kwargs.get("ddtype", None)

    def __getitem__(self, key: Any) -> Any:
        try:
            value = super().__getitem__(key)
            return value
        except KeyError as e:
            if self._ddtype:
                return self._ddtype()
            raise e


@serializable(canonical_name="DictStorePartition", version=1)
class DictStorePartition(KeyValueStorePartition):
    """Dictionary-based StorePartition

    Parameters:
        `settings`: PartitionSettings
            PySyft specific settings, used for indexing and partitioning
        `store_config`: DictStoreConfig
            DictStore specific configuration
    """

    def prune(self) -> None:
        self.init_store()


# the base document store is already a dict but we can change it later
@serializable(canonical_name="DictDocumentStore", version=1)
class DictDocumentStore(DocumentStore):
    """Dictionary-based Document Store

    Parameters:
        `store_config`: DictStoreConfig
            Dictionary Store specific configuration, containing the store type and the backing store type
    """

    partition_type = DictStorePartition

    def __init__(
        self,
        server_uid: uid,
        root_verify_key: SyftVerifyKey | None,
        store_config: DictStoreConfig | None = None,
    ) -> None:
        if store_config is None:
            store_config = DictStoreConfig()
        super().__init__(
            server_uid=server_uid,
            root_verify_key=root_verify_key,
            store_config=store_config,
        )

    def reset(self) -> None:
        for partition in self.partitions.values():
            partition.prune()


@serializable()
class DictStoreConfig(StoreConfig):
    __canonical_name__ = "DictStoreConfig"
    """Dictionary-based configuration

    Parameters:
        `store_type`: Type[DocumentStore]
            The Document type used. Default: DictDocumentStore
        `backing_store`: Type[KeyValueBackingStore]
            The backend type used. Default: DictBackingStore
        locking_config: LockingConfig
            The config used for store locking. Available options:
                * NoLockingConfig: no locking, ideal for single-thread stores.
                * ThreadingLockingConfig: threading-based locking, ideal for same-process in-memory stores.
            Defaults to ThreadingLockingConfig.
    """

    store_type: type[DocumentStore] = DictDocumentStore
    backing_store: type[KeyValueBackingStore] = DictBackingStore
    locking_config: LockingConfig = Field(default_factory=ThreadingLockingConfig)



__version__ = "0.8.7-beta.16"

# stdlib
from collections.abc import Callable
import pathlib
from pathlib import Path
import sys
from typing import Any

# relative
from .abstract_server import ServerSideType
from .abstract_server import ServerType
from .client.client import connect
from .client.client import login
from .client.client import login_as_guest
from .client.client import register
from .client.datasite_client import DatasiteClient
from .client.gateway_client import GatewayClient
from .client.registry import DatasiteRegistry
from .client.registry import EnclaveRegistry
from .client.registry import NetworkRegistry
from .client.search import Search
from .client.search import SearchResults
from .client.syncing import compare_clients
from .client.syncing import compare_states
from .client.syncing import sync
from .client.user_settings import UserSettings
from .client.user_settings import settings
from .custom_worker.config import DockerWorkerConfig
from .custom_worker.config import PrebuiltWorkerConfig
from .orchestra import Orchestra as orchestra
from .protocol.data_protocol import bump_protocol_version
from .protocol.data_protocol import check_or_stage_protocol
from .protocol.data_protocol import get_data_protocol
from .protocol.data_protocol import stage_protocol_changes
from .serde import NOTHING
from .serde.deserialize import _deserialize as deserialize
from .serde.serializable import serializable
from .serde.serialize import _serialize as serialize
from .server.credentials import SyftSigningKey
from .server.datasite import Datasite
from .server.enclave import Enclave
from .server.gateway import Gateway
from .server.uvicorn import serve_server
from .server.uvicorn import serve_server as bind_worker
from .server.worker import Worker
from .service.action.action_data_empty import ActionDataEmpty
from .service.action.action_object import ActionObject
from .service.action.plan import Plan
from .service.action.plan import planify
from .service.api.api import api_endpoint
from .service.api.api import api_endpoint_method
from .service.api.api import create_new_api_endpoint as TwinAPIEndpoint
from .service.code.user_code import UserCodeStatus
from .service.code.user_code import syft_function
from .service.code.user_code import syft_function_single_use
from .service.data_subject import DataSubjectCreate as DataSubject
from .service.dataset.dataset import Contributor
from .service.dataset.dataset import CreateAsset as Asset
from .service.dataset.dataset import CreateDataset as Dataset
from .service.notification.notifications import NotificationStatus
from .service.policy.policy import CreatePolicyRuleConstant as Constant
from .service.policy.policy import CustomInputPolicy
from .service.policy.policy import CustomOutputPolicy
from .service.policy.policy import ExactMatch
from .service.policy.policy import MixedInputPolicy
from .service.policy.policy import SingleExecutionExactOutput
from .service.policy.policy import UserInputPolicy
from .service.policy.policy import UserOutputPolicy
from .service.project.project import ProjectSubmit as Project
from .service.request.request import SubmitRequest as Request
from .service.response import SyftError
from .service.response import SyftNotReady
from .service.response import SyftSuccess
from .service.user.roles import Roles as roles
from .service.user.user_service import UserService
from .stable_version import LATEST_STABLE_SYFT
from .types.twin_object import TwinObject
from .types.uid import UID
from .util import filterwarnings
from .util import options
from .util.autoreload import disable_autoreload
from .util.autoreload import enable_autoreload
from .util.patch_ipython import patch_ipython
from .util.telemetry import instrument
from .util.util import autocache
from .util.util import get_nb_secrets
from .util.util import get_root_data_path
from .util.version_compare import make_requires

requires = make_requires(LATEST_STABLE_SYFT, __version__)


# SYFT_PATH = path = os.path.abspath(a_module.__file__)
SYFT_PATH = pathlib.Path(__file__).parent.resolve()

sys.path.append(str(Path(__file__)))


patch_ipython()


def module_property(func: Any) -> Callable:
    """Decorator to turn module functions into properties.
    Function names must be prefixed with an underscore."""
    module = sys.modules[func.__module__]

    def base_getattr(name: str) -> None:
        raise AttributeError(f"module {module.__name__!r} has no attribute {name!r}")

    old_getattr = getattr(module, "__getattr__", base_getattr)

    def new_getattr(name: str) -> Any:
        if f"_{name}" == func.__name__:
            return func()
        else:
            return old_getattr(name)

    module.__getattr__ = new_getattr  # type: ignore
    return func


@module_property
def _gateways() -> NetworkRegistry:
    return NetworkRegistry()


@module_property
def _enclaves() -> EnclaveRegistry:
    return EnclaveRegistry()


@module_property
def _datasites() -> DatasiteRegistry:
    return DatasiteRegistry()


@module_property
def _settings() -> UserSettings:
    return settings


@module_property
def hello_baby() -> None:
    print("Hello baby!")
    print("Welcome to the world. \u2764\ufe0f")


def search(name: str) -> SearchResults:
    return Search(_datasites()).search(name=name)



# stdlib
from functools import cached_property
import os.path
from pathlib import Path
from typing import Any

# relative
from .builder_docker import DockerBuilder
from .builder_k8s import KubernetesBuilder
from .builder_types import BuilderBase
from .builder_types import ImageBuildResult
from .builder_types import ImagePushResult
from .config import CustomWorkerConfig
from .config import DockerWorkerConfig
from .config import WorkerConfig
from .k8s import IN_KUBERNETES

__all__ = ["CustomWorkerBuilder"]


class CustomWorkerBuilder:
    TYPE_CPU = "cpu"
    TYPE_GPU = "gpu"

    TEMPLATE_DIR_PROD = os.path.expandvars("$APPDIR/grid/images/")
    TEMPLATE_DIR_DEV = "../../../../../grid/backend/grid/images/"

    CUSTOM_IMAGE_PREFIX = "custom-worker"

    BUILD_MAX_WAIT = 30 * 60

    @cached_property
    def builder(self) -> BuilderBase:
        if IN_KUBERNETES:
            return KubernetesBuilder()
        else:
            return DockerBuilder()

    def build_image(
        self,
        config: WorkerConfig,
        tag: str | None = None,
        **kwargs: Any,
    ) -> ImageBuildResult:
        """
        Builds a Docker image from the given configuration.
        Args:
            config (WorkerConfig): The configuration for building the Docker image.
            tag (str): The tag to use for the image.
        """

        if isinstance(config, DockerWorkerConfig):
            return self._build_dockerfile(config, tag, **kwargs)
        elif isinstance(config, CustomWorkerConfig):
            return self._build_template(config, **kwargs)
        else:
            raise TypeError("Unknown worker config type")

    def push_image(
        self,
        tag: str,
        registry_url: str,
        username: str,
        password: str,
        **kwargs: Any,
    ) -> ImagePushResult:
        """
        Pushes a Docker image to the given repo.
        Args:
            repo (str): The repo to push the image to.
            tag (str): The tag to use for the image.
        """

        return self.builder.push_image(
            tag=tag,
            username=username,
            password=password,
            registry_url=registry_url,
        )

    def _build_dockerfile(
        self,
        config: DockerWorkerConfig,
        tag: str,
        **kwargs: Any,
    ) -> ImageBuildResult:
        return self.builder.build_image(
            dockerfile=config.dockerfile,
            tag=tag,
            **kwargs,
        )

    def _build_template(
        self,
        config: CustomWorkerConfig,
        **kwargs: Any,
    ) -> ImageBuildResult:
        # Builds a Docker pre-made CPU/GPU image template using a CustomWorkerConfig
        # remove once GPU is supported
        if config.build.gpu:
            raise Exception("GPU custom worker is not supported yet")

        type = self.TYPE_GPU if config.build.gpu else self.TYPE_CPU

        dockerfile_path = self.find_worker_image(type)

        imgtag = config.get_signature()[:8]

        build_args = {
            "PYTHON_VERSION": config.build.python_version,
            "SYSTEM_PACKAGES": config.build.merged_system_pkgs(),
            "PIP_PACKAGES": config.build.merged_python_pkgs(),
            "CUSTOM_CMD": config.build.merged_custom_cmds(),
        }

        return self.builder.build_image(
            tag=f"{self.CUSTOM_IMAGE_PREFIX}-{type}:{imgtag}",
            dockerfile_path=dockerfile_path,
            buildargs=build_args,
        )

    def find_worker_image(self, type: str) -> Path:
        """
        Find the Worker Dockerfile and it's context path
        - PROD will be in `$APPDIR/grid/`
        - DEV will be in `packages/grid/backend/grid/images`
        - In both the cases context dir does not matter (unless we're calling COPY)

        Args:
            type (str): The type of worker.
        Returns:
            Path: The path to the Dockerfile.
        """
        filename = f"worker_{type}.dockerfile"
        lookup_paths = [
            Path(self.TEMPLATE_DIR_PROD, filename).resolve(),
            Path(__file__, self.TEMPLATE_DIR_DEV, filename).resolve(),
        ]
        for path in lookup_paths:
            if path.exists():
                return path

        raise FileNotFoundError(f"Cannot find the {filename}")



# third party
from typing_extensions import Self

# relative
from ...serde.serializable import serializable
from ...types.base import SyftBaseModel
from ...util.util import get_env
from .server_peer import ServerPeer


def get_rathole_port() -> int:
    return int(get_env("RATHOLE_PORT", "2333"))


@serializable(canonical_name="RatholeConfig", version=1)
class RatholeConfig(SyftBaseModel):
    uuid: str
    secret_token: str
    local_addr_host: str
    local_addr_port: int
    server_name: str | None = None

    @property
    def local_address(self) -> str:
        return f"{self.local_addr_host}:{self.local_addr_port}"

    @classmethod
    def from_peer(cls, peer: ServerPeer) -> Self:
        # relative
        from .routes import HTTPServerRoute

        high_priority_route = peer.pick_highest_priority_route()

        if not isinstance(high_priority_route, HTTPServerRoute):
            raise ValueError("Rathole only supports HTTPServerRoute")

        return cls(
            uuid=peer.id,
            secret_token=peer.rtunnel_token,
            local_addr_host=high_priority_route.host_or_ip,
            local_addr_port=high_priority_route.port,
            server_name=peer.name,
        )



# stdlib
from _ast import Module
import ast
from typing import Any


class GlobalsVisitor(ast.NodeVisitor):
    def generic_visit(self, node: Any) -> None:
        if isinstance(node, ast.Global):
            raise Exception("No Globals allowed!")
        ast.NodeVisitor.generic_visit(self, node)


class LaunchJobVisitor(ast.NodeVisitor):
    def visit_Module(self, node: Module) -> Any:
        self.nested_calls: list = []
        self.generic_visit(node)

    def visit_Call(self, node: Any) -> None:
        if isinstance(node.func, ast.Attribute):
            if (
                getattr(node.func.value, "id", None) == "datasite"
                and node.func.attr == "launch_job"
            ):
                self.nested_calls.append(node.args[0].id)



# stdlib
import logging
import traceback

# third party
from pydantic import EmailStr
from result import Err
from result import Ok
from result import Result

# relative
from ...abstract_server import AbstractServer
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ..context import AuthedServiceContext
from ..notification.notifications import Notification
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from .notifier import NotificationPreferences
from .notifier import NotifierSettings
from .notifier_enums import NOTIFIERS
from .notifier_stash import NotifierStash

logger = logging.getLogger(__name__)


@serializable(canonical_name="NotifierService", version=1)
class NotifierService(AbstractService):
    store: DocumentStore
    stash: NotifierStash  # Which stash should we use?

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = NotifierStash(store=store)

    def settings(  # Maybe just notifier.settings
        self,
        context: AuthedServiceContext,
    ) -> NotifierSettings | SyftError:
        """Get Notifier Settings

        Args:
            context: The request context
        Returns:
            Union[NotifierSettings, SyftError]: Notifier Settings or SyftError
        """
        result = self.stash.get(credentials=context.credentials)
        if result.is_err():
            return SyftError(message="Error getting notifier settings")

        return result.ok()

    def user_settings(
        self,
        context: AuthedServiceContext,
    ) -> NotificationPreferences:
        user_service = context.server.get_service("userservice")
        user_view = user_service.get_current_user(context)
        notifications = user_view.notifications_enabled
        return NotificationPreferences(
            email=notifications[NOTIFIERS.EMAIL],
            sms=notifications[NOTIFIERS.SMS],
            slack=notifications[NOTIFIERS.SLACK],
            app=notifications[NOTIFIERS.APP],
        )

    def turn_on(
        self,
        context: AuthedServiceContext,
        email_username: str | None = None,
        email_password: str | None = None,
        email_sender: str | None = None,
        email_server: str | None = None,
        email_port: int | None = 587,
    ) -> SyftSuccess | SyftError:
        """Turn on email notifications.

        Args:
            email_username (Optional[str]): Email server username. Defaults to None.
            email_password (Optional[str]): Email email server password. Defaults to None.
            sender_email (Optional[str]): Email sender email. Defaults to None.
        Returns:
            Union[SyftSuccess, SyftError]: A union type representing the success or error response.

        Raises:
            None

        """

        result = self.stash.get(credentials=context.credentials)

        # 1 -  If something went wrong at db level, return the error
        if result.is_err():
            return SyftError(message=result.err())

        # 2 - If one of the credentials are set alone, return an error
        if (
            email_username
            and not email_password
            or email_password
            and not email_username
        ):
            return SyftError(message="You must provide both username and password")

        notifier = result.ok()

        # 3 - If notifier doesn't have a email server / port and the user didn't provide them, return an error
        if not (email_server and email_port) and not notifier.email_server:
            return SyftError(
                message="You must provide both server and port to enable notifications."
            )

        logging.debug("Got notifier from db")
        # If no new credentials provided, check for existing ones
        if not (email_username and email_password):
            if not (notifier.email_username and notifier.email_password):
                return SyftError(
                    message="No valid token has been added to the datasite."
                    + "You can add a pair of SMTP credentials via "
                    + "<client>.settings.enable_notifications(email=<>, password=<>)"
                )
            else:
                logging.debug("No new credentials provided. Using existing ones.")
                email_password = notifier.email_password
                email_username = notifier.email_username

        validation_result = notifier.validate_email_credentials(
            username=email_username,
            password=email_password,
            server=email_server if email_server else notifier.email_server,
            port=email_port if email_port else notifier.email_port,
        )

        if validation_result.is_err():
            logging.error(f"Invalid SMTP credentials {validation_result.err()}")
            return SyftError(
                message="Invalid SMTP credentials. Please check your username and password."
            )

        notifier.email_password = email_password
        notifier.email_username = email_username

        if email_server:
            notifier.email_server = email_server
        if email_port:
            notifier.email_port = email_port

        # Email sender verification
        if not email_sender and not notifier.email_sender:
            return SyftError(
                message="You must provide a sender email address to enable notifications."
            )

        if email_sender:
            try:
                EmailStr._validate(email_sender)
            except ValueError:
                return SyftError(
                    message="Invalid sender email address. Please check your email address."
                )
            notifier.email_sender = email_sender

        notifier.active = True
        logging.debug(
            "Email credentials are valid. Updating the notifier settings in the db."
        )

        result = self.stash.update(credentials=context.credentials, settings=notifier)
        if result.is_err():
            return SyftError(message=result.err())
        return SyftSuccess(message="Notifications enabled successfully.")

    def turn_off(
        self,
        context: AuthedServiceContext,
    ) -> SyftSuccess | SyftError:
        """
        Turn off email notifications service.
        PySyft notifications will still work.
        """

        result = self.stash.get(credentials=context.credentials)

        if result.is_err():
            return SyftError(message=result.err())

        notifier = result.ok()
        notifier.active = False
        result = self.stash.update(credentials=context.credentials, settings=notifier)
        if result.is_err():
            return SyftError(message=result.err())
        return SyftSuccess(message="Notifications disabled succesfullly")

    def activate(
        self, context: AuthedServiceContext, notifier_type: NOTIFIERS = NOTIFIERS.EMAIL
    ) -> SyftSuccess | SyftError:
        """
        Activate email notifications for the authenticated user.
        This will only work if the datasite owner has enabled notifications.
        """

        user_service = context.server.get_service("userservice")
        return user_service.enable_notifications(context, notifier_type=notifier_type)

    def deactivate(
        self, context: AuthedServiceContext, notifier_type: NOTIFIERS = NOTIFIERS.EMAIL
    ) -> SyftSuccess | SyftError:
        """Deactivate email notifications for the authenticated user
        This will only work if the datasite owner has enabled notifications.
        """

        user_service = context.server.get_service("userservice")
        return user_service.disable_notifications(context, notifier_type=notifier_type)

    @staticmethod
    def init_notifier(
        server: AbstractServer,
        email_username: str | None = None,
        email_password: str | None = None,
        email_sender: str | None = None,
        smtp_port: int | None = None,
        smtp_host: str | None = None,
    ) -> Result[Ok, Err]:
        """Initialize Notifier settings for a Server.
        If settings already exist, it will use the existing one.
        If not, it will create a new one.

        Args:
            server: Server to initialize the notifier
            active: If notifier should be active
            email_username: Email username to send notifications
            email_password: Email password to send notifications
        Raises:
            Exception: If something went wrong
        Returns:
            Union: SyftSuccess or SyftError
        """
        try:
            # Create a new NotifierStash since its a static method.
            notifier_stash = NotifierStash(store=server.document_store)
            result = notifier_stash.get(server.signing_key.verify_key)
            if result.is_err():
                raise Exception(f"Could not create notifier: {result}")

            # Get the notifier
            notifier = result.ok()
            # If notifier doesn't exist, create a new one

            if not notifier:
                notifier = NotifierSettings()
                notifier.active = False  # Default to False

            # TODO: this should be a method in NotifierSettings
            if email_username and email_password:
                validation_result = notifier.validate_email_credentials(
                    username=email_username,
                    password=email_password,
                    server=smtp_host,
                    port=smtp_port,
                )

                sender_not_set = not email_sender and not notifier.email_sender
                if validation_result.is_err() or sender_not_set:
                    logger.error(
                        f"Notifier validation error - {validation_result.err()}.",
                    )
                    notifier.active = False
                else:
                    notifier.email_password = email_password
                    notifier.email_username = email_username
                    notifier.email_sender = email_sender
                    notifier.email_server = smtp_host
                    notifier.email_port = smtp_port
                    notifier.active = True

            notifier_stash.set(server.signing_key.verify_key, notifier)
            return Ok("Notifier initialized successfully")

        except Exception:
            raise Exception(f"Error initializing notifier. \n {traceback.format_exc()}")

    # This is not a public API.
    # This method is used by other services to dispatch notifications internally
    def dispatch_notification(
        self, context: AuthedServiceContext, notification: Notification
    ) -> SyftError:
        admin_key = context.server.get_service("userservice").admin_verify_key()
        notifier = self.stash.get(admin_key)
        if notifier.is_err():
            return SyftError(
                message="The mail service ran out of quota or some notifications failed to be delivered.\n"
                + "Please check the health of the mailing server."
            )

        notifier = notifier.ok()
        # If notifier is active
        if notifier.active:
            resp = notifier.send_notifications(
                context=context, notification=notification
            )
            if resp.is_err():
                return SyftError(message=resp.err())

        # If notifier isn't active, return None
        return SyftSuccess(message="Notifications dispatched successfully")



# relative
from ...types.server_url import ServerURL
from .rathole_config_builder import RatholeConfigBuilder
from .routes import ServerRoute
from .server_peer import ServerPeer


class ReverseTunnelService:
    def __init__(self) -> None:
        self.builder = RatholeConfigBuilder()

    def set_client_config(
        self,
        self_server_peer: ServerPeer,
        remote_server_route: ServerRoute,
    ) -> None:
        rathole_route = self_server_peer.get_rtunnel_route()
        if not rathole_route:
            raise Exception(
                "Failed to exchange routes via . "
                + f"Peer: {self_server_peer} has no rathole route: {rathole_route}"
            )

        remote_url = ServerURL(
            host_or_ip=remote_server_route.host_or_ip, port=remote_server_route.port
        )
        rathole_remote_addr = remote_url.as_container_host()

        remote_addr = rathole_remote_addr.url_no_protocol

        self.builder.add_host_to_client(
            peer_name=self_server_peer.name,
            peer_id=str(self_server_peer.id),
            rtunnel_token=rathole_route.rtunnel_token,
            remote_addr=remote_addr,
        )

    def set_server_config(self, remote_peer: ServerPeer) -> None:
        rathole_route = remote_peer.get_rtunnel_route()
        self.builder.add_host_to_server(remote_peer) if rathole_route else None

    def clear_client_config(self, self_server_peer: ServerPeer) -> None:
        self.builder.remove_host_from_client(str(self_server_peer.id))

    def clear_server_config(self, remote_peer: ServerPeer) -> None:
        self.builder.remove_host_from_server(
            str(remote_peer.id), server_name=remote_peer.name
        )



# stdlib
from collections import defaultdict
import json
import os
from pathlib import Path
from typing import Any

# syft absolute
import syft as sy

# relative
from .decorators import deprecated

RELATIVE_PATH_TO_FRONTEND = "/../../../../grid/frontend/"
SCHEMA_FOLDER = "schema"

GUEST_COMMANDS = """
<li><span class='syft-code-block'>&lt;your_client&gt;.datasets</span> - list datasets</li>
<li><span class='syft-code-block'>&lt;your_client&gt;.code</span> - list code</li>
<li><span class='syft-code-block'>&lt;your_client&gt;.login</span> - list projects</li>
"""

DS_COMMANDS = """
<li><span class='syft-code-block'>&lt;your_client&gt;.datasets</span> - list datasets</li>
<li><span class='syft-code-block'>&lt;your_client&gt;.code</span> - list code</li>
<li><span class='syft-code-block'>&lt;your_client&gt;.projects</span> - list projects</li>
"""

DO_COMMANDS = """
<li><span class='syft-code-block'>&lt;your_client&gt;.projects</span> - list projects</li>
<li><span class='syft-code-block'>&lt;your_client&gt;.requests</span> - list requests</li>
<li><span class='syft-code-block'>&lt;your_client&gt;.users</span> - list users</li>
"""

DEFAULT_WELCOME_MSG = """
        <style>
            $FONT_CSS

            .syft-container {
                padding: 5px;
                font-family: 'Open Sans';
            }
            .syft-alert-info {
                color: #1F567A;
                background-color: #C2DEF0;
                border-radius: 4px;
                padding: 5px;
                padding: 13px 10px
            }
            .syft-code-block {
                background-color: #f7f7f7;
                border: 1px solid #cfcfcf;
                padding: 0px 2px;
            }
            .syft-space {
                margin-top: 1em;
            }
        </style>
        <div class="syft-client syft-container">
            <img src="$server_symbol" alt="Logo"
            style="width:48px;height:48px;padding:3px;">
            <h2>Welcome to $datasite_name</h2>
            <div class="syft-space">
            <strong>URL:</strong> $server_url <br />
            <strong>Server Description:</strong> $description <br />
            <strong>Server Type:</strong> $server_type <br />
            <strong>Server Side Type:</strong>$server_side_type<br />
            <strong>Syft Version:</strong> $server_version<br />

            </div>
            <div class='syft-alert-info syft-space'>
                &#9432;&nbsp;
                This datasite is run by the library PySyft to learn more about how it works visit
                <a href="https://github.com/OpenMined/PySyft">github.com/OpenMined/PySyft</a>.
            </div>
            <h4>Commands to Get Started</h4>
            $command_list
        </div><br />
        """

# json schema primitive types
primitive_mapping = {
    list: "array",
    bool: "boolean",
    int: "integer",
    type(None): "null",
    float: "number",
    dict: "object",
    str: "string",
}


def make_fake_type(_type_str: str) -> dict[str, Any]:
    jsonschema: dict[str, Any] = {}
    jsonschema["title"] = _type_str
    jsonschema["type"] = "object"
    jsonschema["properties"] = {}
    jsonschema["additionalProperties"] = False
    return jsonschema


def get_type_mapping(_type: type) -> str:
    if _type in primitive_mapping:
        return primitive_mapping[_type]
    return _type.__name__


def get_types(cls: type, keys: list[str]) -> dict[str, type] | None:
    types = []
    for key in keys:
        _type = None
        if key in cls.__annotations__:
            _type = cls.__annotations__[key]
        else:
            for parent_cls in cls.mro():
                annotations = getattr(parent_cls, "__annotations__", None)
                if annotations and key in annotations:
                    _type = annotations[key]
        if _type is None:
            # print(f"Failed to find type for key: {key} in {cls}")
            return None
        types.append(_type)
    return dict(zip(keys, types))


def convert_attribute_types(
    cls: type, attribute_list: list[str], attribute_types: list[type]
) -> dict[str, Any]:
    jsonschema: dict[str, Any] = {}
    jsonschema["title"] = cls.__name__
    jsonschema["type"] = "object"
    jsonschema["properties"] = {}
    jsonschema["additionalProperties"] = False
    for attribute, _type in dict(zip(attribute_list, attribute_types)).items():
        attribute_dict = {}
        attribute_dict["type"] = get_type_mapping(_type)
        jsonschema["properties"][attribute] = attribute_dict
    jsonschema["required"] = list(attribute_list)
    return jsonschema


def process_type_bank(type_bank: dict[str, tuple[Any, ...]]) -> dict[str, dict]:
    # first pass gets each type into basic json schema format
    json_mappings = {}
    count = 0
    converted_types: dict[str, int] = defaultdict(int)
    for k in type_bank:
        count += 1
        t = type_bank[k]
        (
            nonrecursive,
            serialize,
            deserialize,
            attribute_list,
            exclude_attrs,
            serde_overrides,
            cls,
            attribute_types,
        ) = t

        lib = cls.__module__.split(".")[0]

        # process types with an attribute list and attribute types first
        if attribute_list and attribute_types:
            if ".uid." not in str(cls):
                print(f"Skipping {k}")
                continue
            try:
                print(f"Processing {k}")
                schema = convert_attribute_types(cls, attribute_list, attribute_types)
                json_mappings[cls.__name__] = schema
                converted_types[lib] += 1
            except Exception:  # nosec
                print(f"Failed to process. {k}")

        # TODO: process other types of serializable objects

    converted = sum(converted_types.values())
    print(f"Finished converting {converted} out of {count}")
    return json_mappings


def resolve_references(json_mappings: dict[str, dict]) -> dict[str, dict]:
    # track second pass generated types
    new_types = {}
    for json_schema in json_mappings.values():
        replace_types = {}
        for attribute, config in json_schema["properties"].items():
            if "type" in config:
                _type_str = config["type"]
                if _type_str in primitive_mapping.values():
                    # no issue with primitive types
                    continue
                else:
                    # if we don't have a type yet its because its not supported
                    # lets create an empty type to satisfy the generation process
                    if _type_str not in json_mappings.keys():
                        reference = make_fake_type(_type_str)
                        new_types[_type_str] = reference

                # if the type is a reference we need to replace its type entry after
                replace_types[attribute] = {
                    "$ref": f"./{SCHEMA_FOLDER}/{_type_str}.json"
                }
        # replace any referenced types
        for k, v in replace_types.items():
            json_schema["properties"][k] = v

    # insert any new types created above into the main dict
    for k, v in new_types.items():
        json_mappings[k] = v

    return json_mappings


@deprecated(
    reason="generate_json_schemas is outdated, #1603 for more info",
)
def generate_json_schemas(output_path: str | None = None) -> None:
    # TODO: should we also replace this with the SyftObjectRegistry?
    json_mappings = process_type_bank(sy.serde.recursive.TYPE_BANK)
    json_mappings = resolve_references(json_mappings)
    if not output_path:
        output_path = os.path.dirname(__file__) + RELATIVE_PATH_TO_FRONTEND
        output_path = os.path.abspath(output_path)
    path = Path(output_path) / SCHEMA_FOLDER
    os.makedirs(path, exist_ok=True)
    for k, v in json_mappings.items():
        with open(path / f"{k}.json", "w") as f:
            f.write(json.dumps(v))



# stdlib
from typing import Any

# third party
from bson import CodecOptions
from bson.binary import Binary
from bson.binary import USER_DEFINED_SUBTYPE
from bson.codec_options import TypeDecoder
from bson.codec_options import TypeRegistry

# relative
from ..serde.deserialize import _deserialize
from ..serde.serialize import _serialize


def fallback_syft_encoder(value: object) -> Binary:
    return Binary(_serialize(value, to_bytes=True), USER_DEFINED_SUBTYPE)


class SyftMongoBinaryDecoder(TypeDecoder):
    bson_type = Binary

    def transform_bson(self, value: Any) -> Any:
        if value.subtype == USER_DEFINED_SUBTYPE:
            return _deserialize(value, from_bytes=True)
        return value


syft_codecs = [SyftMongoBinaryDecoder()]
syft_type_registry = TypeRegistry(syft_codecs, fallback_encoder=fallback_syft_encoder)
SYFT_CODEC_OPTIONS = CodecOptions(type_registry=syft_type_registry)



# stdlib

# third party
from result import Ok
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionKey
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ..response import SyftSuccess
from .image_registry import SyftImageRegistry

__all__ = ["SyftImageRegistryStash"]


URLPartitionKey = PartitionKey(key="url", type_=str)


@serializable(canonical_name="SyftImageRegistryStash", version=1)
class SyftImageRegistryStash(BaseUIDStoreStash):
    object_type = SyftImageRegistry
    settings: PartitionSettings = PartitionSettings(
        name=SyftImageRegistry.__canonical_name__,
        object_type=SyftImageRegistry,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_by_url(
        self,
        credentials: SyftVerifyKey,
        url: str,
    ) -> Result[SyftImageRegistry | None, str]:
        qks = QueryKeys(qks=[URLPartitionKey.with_obj(url)])
        return self.query_one(credentials=credentials, qks=qks)

    def delete_by_url(
        self, credentials: SyftVerifyKey, url: str
    ) -> Result[SyftSuccess, str]:
        qk = URLPartitionKey.with_obj(url)
        result = super().delete(credentials=credentials, qk=qk)
        if result.is_ok():
            return Ok(SyftSuccess(message=f"URL: {url} deleted"))
        return result



# stdlib
from enum import Enum


class Roles(Enum):
    SOURCE_CONTRIBUTOR = "Source Contributor"
    UPLOADER = "Uploader"
    EDITOR = "Editor"

    def __str__(self) -> str:
        return f"{self.value}"



# stdlib
import json
from typing import Any

# relative
from ...client.api import APIRegistry
from ...serde.serializable import serializable
from ...service.user.user_roles import ServiceRole
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.syft_object import SyftVerifyKey
from ...types.uid import UID
from ...util.notebook_ui.components.tabulator_template import (
    build_tabulator_table_with_data,
)
from ...util.table import prepare_table_data
from ..code.user_code import UserCode
from ..response import SyftError


@serializable()
class CodeHistory(SyftObject):
    # version
    __canonical_name__ = "CodeHistory"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    server_uid: UID
    user_verify_key: SyftVerifyKey
    user_code_history: list[UID] = []
    service_func_name: str
    comment_history: list[str] = []

    __attr_searchable__ = ["user_verify_key", "service_func_name"]

    def add_code(self, code: UserCode, comment: str | None = None) -> None:
        self.user_code_history.append(code.id)
        if comment is None:
            comment = ""
        self.comment_history.append(comment)


@serializable()
class CodeHistoryView(SyftObject):
    # version
    __canonical_name__ = "CodeHistoryView"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    user_code_history: list[UserCode] = []
    service_func_name: str
    comment_history: list[str] = []

    def _coll_repr_(self) -> dict[str, int]:
        return {"Number of versions": len(self.user_code_history)}

    def _repr_html_(self) -> str | None:
        rows, metadata = prepare_table_data(self.user_code_history)

        for i, r in enumerate(rows):
            r["Version"] = f"v{i}"
            raw_code = self.user_code_history[i].raw_code
            n_code_lines = raw_code.count("\n")
            if n_code_lines > 5:
                raw_code = "\n".join(raw_code.split("\n", 5))
            r["Code"] = raw_code

        metadata["name"] = "Code History"
        metadata["columns"] += ["Version", "Code"]

        return build_tabulator_table_with_data(rows, metadata)

    def __getitem__(self, index: int | str) -> UserCode | SyftError:
        if isinstance(index, str):
            raise TypeError(f"index {index} must be an integer, not a string")
        api = APIRegistry.api_for(
            self.syft_server_location, self.syft_client_verify_key
        )
        if api is None:
            return SyftError(
                message=f"Can't access the api. You must login to {self.server_uid}"
            )
        if (
            api.user.get_current_user().role.value >= ServiceRole.DATA_OWNER.value
            and index < 0
        ):
            # negative index would dynamically resolve to a different version
            return SyftError(
                message="For security concerns we do not allow negative indexing. \
                Try using absolute values when indexing"
            )
        return self.user_code_history[index]


@serializable()
class CodeHistoriesDict(SyftObject):
    # version
    __canonical_name__ = "CodeHistoriesDict"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    code_versions: dict[str, CodeHistoryView] = {}

    def _repr_html_(self) -> str:
        return f"""
            {self.code_versions._repr_html_()}
            """

    def add_func(self, versions: CodeHistoryView) -> Any:
        self.code_versions[versions.service_func_name] = versions

    def __getitem__(self, name: str | int) -> Any:
        if isinstance(name, int):
            raise TypeError("name argument ({name}) must be a string, not an integer.")
        return self.code_versions[name]

    def __getattr__(self, name: str) -> Any:
        code_versions = object.__getattribute__(self, "code_versions")
        if name in code_versions.keys():
            return code_versions[name]
        return object.__getattribute__(self, name)


@serializable()
class UsersCodeHistoriesDict(SyftObject):
    # version
    __canonical_name__ = "UsersCodeHistoriesDict"
    __version__ = SYFT_OBJECT_VERSION_1

    id: UID
    server_uid: UID
    user_dict: dict[str, list[str]] = {}

    __repr_attrs__ = ["available_keys"]

    @property
    def available_keys(self) -> str:
        return json.dumps(self.user_dict, sort_keys=True, indent=4)

    def __getitem__(self, key: str | int) -> CodeHistoriesDict | SyftError:
        api = APIRegistry.api_for(self.server_uid, self.syft_client_verify_key)
        if api is None:
            return SyftError(
                message=f"Can't access the api. You must login to {self.server_uid}"
            )
        return api.services.code_history.get_history_for_user(key)

    def _repr_html_(self) -> str | None:
        rows = [
            {"User": user, "UserCodes": ", ".join(funcs)}
            for user, funcs in self.user_dict.items()
        ]
        metadata = {
            "name": "UserCode Histories",
            "columns": ["User", "UserCodes"],
            "icon": None,
        }
        return build_tabulator_table_with_data(rows, metadata)



# stdlib
from collections.abc import Iterable
import json


def iterator_to_string(iterator: Iterable) -> str:
    log = ""
    for line in iterator:
        for item in line.values():
            if isinstance(item, str):
                log += item
            elif isinstance(item, dict):
                log += json.dumps(item) + "\n"
            else:
                log += str(item)
    return log


class ImageUtils:
    @staticmethod
    def parse_tag(tag: str) -> tuple[str | None, str, str]:
        url, tag = tag.rsplit(":", 1)
        args = url.rsplit("/", 2)

        if len(args) == 3:
            registry = args[0]
            repo = "/".join(args[1:])
        else:
            registry = None
            repo = "/".join(args)

        return registry, repo, tag

    @staticmethod
    def change_registry(tag: str, registry: str) -> str:
        _, repo, tag = ImageUtils.parse_tag(tag)
        return f"{registry}/{repo}:{tag}"



"""
Capture stdout, stderr and stdin streams

References:
- https://github.com/OpenMined/PySyft/pull/8560
- https://github.com/pytest-dev/py/blob/master/py/_io/capture.py
"""

# stdlib
from collections.abc import Callable
from collections.abc import Generator
import contextlib
import os
import sys
import tempfile
from typing import Any
from typing import cast

patchsysdict = {0: "stdin", 1: "stdout", 2: "stderr"}

try:
    devnullpath = os.devnull
except AttributeError:
    if os.name == "nt":
        devnullpath = "NUL"
    else:
        devnullpath = "/dev/null"


class DontReadFromInput:
    """Temporary stub class.  Ideally when stdin is accessed, the
    capturing should be turned off, with possibly all data captured
    so far sent to the screen.  This should be configurable, though,
    because in automated test runs it is better to crash than
    hang indefinitely.
    """

    def read(self, *args: Any) -> None:
        raise OSError("reading from stdin while output is captured")

    readline = read
    readlines = read
    __iter__ = read

    def fileno(self) -> None:
        raise ValueError("redirected Stdin is pseudofile, has no fileno()")

    def isatty(self) -> bool:
        return False

    def close(self) -> None:
        pass


class Capture:
    @classmethod
    def call(cls, func: Callable, *args: Any, **kwargs: Any) -> tuple[Any, str, str]:
        """return a (res, out, err) tuple where
        out and err represent the output/error output
        during function execution.
        call the given function with args/kwargs
        and capture output/error during its execution.
        """
        so = cls()
        try:
            res = func(*args, **kwargs)
        finally:
            out, err = so.reset()
        return res, out, err

    def reset(self) -> tuple[str, str]:
        """reset sys.stdout/stderr and return captured output as strings."""
        if hasattr(self, "_reset"):
            raise ValueError("was already reset")
        self._reset = True
        outfile, errfile = self.done(save=False)
        out, err = "", ""
        if outfile and not outfile.closed:
            out = outfile.read()
            outfile.close()
        if errfile and errfile != outfile and not errfile.closed:
            err = errfile.read()
            errfile.close()
        return out, err

    def suspend(self) -> tuple[str, str]:
        """return current snapshot captures, memorize tempfiles."""
        outerr = self.readouterr()
        outfile, errfile = self.done()
        return outerr


class FDCapture:
    """Capture IO to/from a given os-level filedescriptor."""

    def __init__(
        self,
        targetfd: int,
        tmpfile: Any | None = None,
        now: bool = True,
        patchsys: bool = False,
    ) -> None:
        """save targetfd descriptor, and open a new
        temporary file there.  If no tmpfile is
        specified a tempfile.Tempfile() will be opened
        in text mode.
        """
        self.targetfd = targetfd
        if tmpfile is None and targetfd != 0:
            f = tempfile.TemporaryFile("wb+")
            tmpfile = dupfile(f, encoding="UTF-8")
            f.close()
        self.tmpfile = cast(Any, tmpfile)
        self._savefd = os.dup(self.targetfd)
        if patchsys:
            self._oldsys = getattr(sys, patchsysdict[targetfd])
        if now:
            self.start()

    def start(self) -> None:
        try:
            os.fstat(self._savefd)
        except OSError:
            raise ValueError(
                "saved filedescriptor not valid, " "did you call start() twice?"
            )
        if self.targetfd == 0 and not self.tmpfile:
            fd = os.open(devnullpath, os.O_RDONLY)
            os.dup2(fd, 0)
            os.close(fd)
            if hasattr(self, "_oldsys"):
                setattr(sys, patchsysdict[self.targetfd], DontReadFromInput())
        else:
            os.dup2(self.tmpfile.fileno(), self.targetfd)
            if hasattr(self, "_oldsys"):
                setattr(sys, patchsysdict[self.targetfd], self.tmpfile)

    def done(self) -> Any:
        """unpatch and clean up, returns the self.tmpfile (file object)"""
        os.dup2(self._savefd, self.targetfd)
        os.close(self._savefd)
        if self.targetfd != 0:
            self.tmpfile.seek(0)
        if hasattr(self, "_oldsys"):
            setattr(sys, patchsysdict[self.targetfd], self._oldsys)
        return self.tmpfile

    def writeorg(self, data: bytes) -> None:
        """write a string to the original file descriptor"""
        tempfp = tempfile.TemporaryFile()
        try:
            os.dup2(self._savefd, tempfp.fileno())
            tempfp.write(data)
        finally:
            tempfp.close()


class StdCaptureFD(Capture):
    """This class allows to capture writes to FD1 and FD2
    and may connect a NULL file to FD0 (and prevent
    reads from sys.stdin).  If any of the 0,1,2 file descriptors
    is invalid it will not be captured.
    """

    def __init__(
        self,
        out: bool = True,
        err: bool = True,
        mixed: bool = False,
        in_: bool = True,
        patchsys: bool = True,
        now: bool = True,
    ):
        self._options = {
            "out": out,
            "err": err,
            "mixed": mixed,
            "in_": in_,
            "patchsys": patchsys,
            "now": now,
        }
        self._save()
        if now:
            self.startall()

    def _save(self) -> None:
        in_ = self._options["in_"]
        out = self._options["out"]
        err = self._options["err"]
        mixed = self._options["mixed"]
        patchsys = self._options["patchsys"]
        if in_:
            try:
                self.in_ = FDCapture(0, tmpfile=None, now=False, patchsys=patchsys)
            except OSError:
                pass
        if out:
            tmpfile = None
            if hasattr(out, "write"):
                tmpfile = out
            try:
                self.out = FDCapture(1, tmpfile=tmpfile, now=False, patchsys=patchsys)
                self._options["out"] = self.out.tmpfile
            except OSError:
                pass
        if err:
            if out and mixed:
                tmpfile = self.out.tmpfile
            elif hasattr(err, "write"):
                tmpfile = err
            else:
                tmpfile = None
            try:
                self.err = FDCapture(2, tmpfile=tmpfile, now=False, patchsys=patchsys)
                self._options["err"] = self.err.tmpfile
            except OSError:
                pass

    def startall(self) -> None:
        if hasattr(self, "in_"):
            self.in_.start()
        if hasattr(self, "out"):
            self.out.start()
        if hasattr(self, "err"):
            self.err.start()

    def resume(self) -> None:
        """resume capturing with original temp files."""
        self.startall()

    def done(self, save: bool = True) -> tuple[Any | None, Any | None]:
        """return (outfile, errfile) and stop capturing."""
        outfile = errfile = None
        if hasattr(self, "out") and not self.out.tmpfile.closed:
            outfile = self.out.done()
        if hasattr(self, "err") and not self.err.tmpfile.closed:
            errfile = self.err.done()
        if hasattr(self, "in_"):
            self.in_.done()
        if save:
            self._save()
        return outfile, errfile

    def readouterr(self) -> tuple[str, str]:
        """return snapshot value of stdout/stderr capturings."""
        if hasattr(self, "out"):
            out = self._readsnapshot(self.out.tmpfile)
        else:
            out = ""
        if hasattr(self, "err"):
            err = self._readsnapshot(self.err.tmpfile)
        else:
            err = ""
        return out, err

    def _readsnapshot(self, f: Any) -> str:
        f.seek(0)
        res = f.read()
        enc = getattr(f, "encoding", None)
        if enc:

            def _totext(
                obj: Any, encoding: str | None = None, errors: str | None = None
            ) -> str:
                """
                Source: https://github.com/pytest-dev/py/blob/master/py/_builtin.py
                """
                if isinstance(obj, bytes):
                    if errors is None:
                        obj = obj.decode(encoding)
                    else:
                        obj = obj.decode(encoding, errors)
                elif not isinstance(obj, str):
                    obj = str(obj)
                return obj

            res = _totext(res, enc, "replace")
        f.truncate(0)
        f.seek(0)
        return res


def dupfile(
    f: Any,
    mode: str | None = None,
    buffering: int = 0,
    raising: bool = False,
    encoding: str | None = None,
) -> Any:
    """return a new open file object that's a duplicate of f

    mode is duplicated if not given, 'buffering' controls
    buffer size (defaulting to no buffering) and 'raising'
    defines whether an exception is raised when an incompatible
    file object is passed in (if raising is False, the file
    object itself will be returned)
    """
    try:
        fd = f.fileno()
        mode = mode or f.mode
        newfd = os.dup(fd)
    except AttributeError:
        if raising:
            raise
        return f

    if encoding is not None:
        mode = mode.replace("b", "")
        buffering = True

    return os.fdopen(newfd, mode, buffering, encoding, closefd=True)


@contextlib.contextmanager
def std_stream_capture(out: bool = True, err: bool = True) -> Generator[Any, None, Any]:
    try:
        capture = StdCaptureFD(out=out, err=err)
    except Exception:
        capture = None

    try:
        yield
    finally:
        if capture is not None:
            capture.reset()



# future
from __future__ import annotations

# stdlib
from collections import defaultdict
from collections.abc import Callable
from copy import deepcopy
import functools
from functools import partial
import inspect
from inspect import Parameter
import logging
from typing import Any
from typing import TYPE_CHECKING

# third party
from result import Ok
from result import OkErr
from typing_extensions import Self

# relative
from ..abstract_server import AbstractServer
from ..protocol.data_protocol import migrate_args_and_kwargs
from ..serde.lib_permissions import CMPCRUDPermission
from ..serde.lib_permissions import CMPPermission
from ..serde.lib_service_registry import CMPBase
from ..serde.lib_service_registry import CMPClass
from ..serde.lib_service_registry import CMPFunction
from ..serde.lib_service_registry import action_execute_registry_libs
from ..serde.serializable import serializable
from ..serde.signature import Signature
from ..serde.signature import signature_remove_context
from ..serde.signature import signature_remove_self
from ..server.credentials import SyftVerifyKey
from ..store.document_store import DocumentStore
from ..store.linked_obj import LinkedObject
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftBaseObject
from ..types.syft_object import SyftObject
from ..types.syft_object import attach_attribute_to_syft_object
from ..types.uid import UID
from .context import AuthedServiceContext
from .context import ChangeContext
from .response import SyftError
from .user.user_roles import DATA_OWNER_ROLE_LEVEL
from .user.user_roles import ServiceRole
from .warnings import APIEndpointWarning

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    # relative
    from ..client.api import APIModule

TYPE_TO_SERVICE: dict = {}
SERVICE_TO_TYPES: defaultdict = defaultdict(set)


class AbstractService:
    server: AbstractServer
    server_uid: UID
    store_type: type = DocumentStore

    def resolve_link(
        self,
        context: AuthedServiceContext | ChangeContext | Any,
        linked_obj: LinkedObject,
    ) -> Any | SyftError:
        if isinstance(context, AuthedServiceContext):
            credentials = context.credentials
        elif isinstance(context, ChangeContext):
            credentials = context.approving_user_credentials
        else:
            return SyftError(message="wrong context passed")

        obj = self.stash.get_by_uid(credentials, uid=linked_obj.object_uid)
        if isinstance(obj, OkErr) and obj.is_ok():
            obj = obj.ok()
        if hasattr(obj, "server_uid"):
            if context.server is None:
                return SyftError(message=f"context {context}'s server is None")
            obj.server_uid = context.server.id
        if not isinstance(obj, OkErr):
            obj = Ok(obj)
        return obj

    def get_all(*arg: Any, **kwargs: Any) -> Any:
        pass


@serializable()
class BaseConfig(SyftBaseObject):
    __canonical_name__ = "BaseConfig"
    __version__ = SYFT_OBJECT_VERSION_1

    public_path: str
    private_path: str
    public_name: str
    method_name: str
    doc_string: str | None = None
    signature: Signature | None = None
    is_from_lib: bool = False
    warning: APIEndpointWarning | None = None


@serializable()
class ServiceConfig(BaseConfig):
    __canonical_name__ = "ServiceConfig"
    permissions: list
    roles: list[ServiceRole]

    def has_permission(self, user_service_role: ServiceRole) -> bool:
        return user_service_role in self.roles


@serializable()
class LibConfig(BaseConfig):
    __canonical_name__ = "LibConfig"
    permissions: set[CMPPermission]

    def has_permission(self, credentials: SyftVerifyKey) -> bool:
        # TODO: implement user level permissions
        for p in self.permissions:
            if p.permission_string == CMPCRUDPermission.ALL_EXECUTE.name:
                return True
            if p.permission_string == CMPCRUDPermission.NONE_EXECUTE.name:
                return False
        return False


class ServiceConfigRegistry:
    __service_config_registry__: dict[str, ServiceConfig] = {}
    # __public_to_private_path_map__: Dict[str, str] = {}

    @classmethod
    def register(cls, config: ServiceConfig) -> None:
        if not cls.path_exists(config.public_path):
            cls.__service_config_registry__[config.public_path] = config
            # cls.__public_to_private_path_map__[config.public_path] = config.private_path

    @classmethod
    def get_registered_configs(cls) -> dict[str, ServiceConfig]:
        return cls.__service_config_registry__

    @classmethod
    def path_exists(cls, path: str) -> bool:
        return path in cls.__service_config_registry__


class LibConfigRegistry:
    __service_config_registry__: dict[str, ServiceConfig] = {}

    @classmethod
    def register(cls, config: ServiceConfig) -> None:
        if not cls.path_exists(config.public_path):
            cls.__service_config_registry__[config.public_path] = config

    @classmethod
    def get_registered_configs(cls) -> dict[str, ServiceConfig]:
        return cls.__service_config_registry__

    @classmethod
    def path_exists(cls, path: str) -> bool:
        return path in cls.__service_config_registry__


class UserLibConfigRegistry:
    def __init__(self, service_config_registry: dict[str, LibConfig]):
        self.__service_config_registry__: dict[str, LibConfig] = service_config_registry

    @classmethod
    def from_user(cls, credentials: SyftVerifyKey) -> Self:
        return cls(
            {
                k: lib_config
                for k, lib_config in LibConfigRegistry.get_registered_configs().items()
                if lib_config.has_permission(credentials)
            }
        )

    def __contains__(self, path: str) -> bool:
        return path in self.__service_config_registry__

    def private_path_for(self, public_path: str) -> str:
        return self.__service_config_registry__[public_path].private_path

    def get_registered_configs(self) -> dict[str, LibConfig]:
        return self.__service_config_registry__


class UserServiceConfigRegistry:
    def __init__(self, service_config_registry: dict[str, ServiceConfig]):
        self.__service_config_registry__: dict[str, ServiceConfig] = (
            service_config_registry
        )

    @classmethod
    def from_role(cls, user_service_role: ServiceRole) -> Self:
        return cls(
            {
                k: service_config
                for k, service_config in ServiceConfigRegistry.get_registered_configs().items()
                if service_config.has_permission(user_service_role)
            }
        )

    def __contains__(self, path: str) -> bool:
        return path in self.__service_config_registry__

    def private_path_for(self, public_path: str) -> str:
        return self.__service_config_registry__[public_path].private_path

    def get_registered_configs(self) -> dict[str, ServiceConfig]:
        return self.__service_config_registry__


def register_lib_obj(lib_obj: CMPBase) -> None:
    signature = lib_obj.signature
    path = lib_obj.absolute_path
    func_name = lib_obj.name

    if signature is not None:
        if path != "numpy.source":
            lib_config = LibConfig(
                public_path=str(path),
                private_path=str(path),
                public_name=str(func_name),
                method_name=str(func_name),
                doc_string=str(lib_obj.__doc__),
                signature=signature,
                permissions={lib_obj.permissions},
                is_from_lib=True,
            )

            LibConfigRegistry.register(lib_config)


# hacky, prevent circular imports
for lib_obj in action_execute_registry_libs.flatten():
    # # for functions
    # func_name = func.__name__
    # # for classes
    # func_name = path.split(".")[-1]
    if isinstance(lib_obj, CMPFunction) or isinstance(lib_obj, CMPClass):
        register_lib_obj(lib_obj)


def deconstruct_param(param: inspect.Parameter) -> dict[str, Any]:
    # Gets the init signature form pydantic object
    param_type = param.annotation
    if not hasattr(param_type, "__signature__"):
        raise Exception(
            f"Type {param_type} needs __signature__. Or code changed to support backup init"
        )
    signature = param_type.__signature__
    sub_mapping = {}
    for k, v in signature.parameters.items():
        sub_mapping[k] = v
    return sub_mapping


def types_for_autosplat(signature: Signature, autosplat: list[str]) -> dict[str, type]:
    autosplat_types = {}
    for k, v in signature.parameters.items():
        if k in autosplat:
            autosplat_types[k] = v.annotation
    return autosplat_types


def reconstruct_args_kwargs(
    signature: Signature,
    autosplat: list[str],
    args: tuple[Any, ...],
    kwargs: dict[Any, str],
) -> tuple[tuple[Any, ...], dict[str, Any]]:
    autosplat_types = types_for_autosplat(signature=signature, autosplat=autosplat)

    autosplat_objs = {}
    for autosplat_key, autosplat_type in autosplat_types.items():
        init_kwargs = {}
        keys = autosplat_type.__fields__.keys()
        for key in keys:
            if key in kwargs:
                init_kwargs[key] = kwargs.pop(key)
        autosplat_objs[autosplat_key] = autosplat_type(**init_kwargs)

    final_kwargs = {}
    for param_key, param in signature.parameters.items():
        if param_key in kwargs:
            final_kwargs[param_key] = kwargs[param_key]
        elif param_key in autosplat_objs:
            final_kwargs[param_key] = autosplat_objs[param_key]
        elif not isinstance(param.default, type(Parameter.empty)):
            final_kwargs[param_key] = param.default
        else:
            raise Exception(f"Missing {param_key} not in kwargs.")

    if "context" in kwargs:
        final_kwargs["context"] = kwargs["context"]

    return (args, final_kwargs)


def expand_signature(signature: Signature, autosplat: list[str]) -> Signature:
    new_mapping = {}
    for k, v in signature.parameters.items():
        if k in autosplat:
            sub_mapping = deconstruct_param(v)
            for s, t in sub_mapping.items():
                new_t_kwargs = {
                    "annotation": t.annotation,
                    "name": t.name,
                    "default": t.default,
                    "kind": Parameter.POSITIONAL_OR_KEYWORD,
                }
                new_t = Parameter(**new_t_kwargs)
                new_mapping[s] = new_t
        else:
            new_mapping[k] = v

    # Reorder the parameter based on if they have default value or not
    new_params = sorted(
        new_mapping.values(),
        key=lambda param: param.default is param.empty,
        reverse=True,
    )

    return Signature(
        **{
            "parameters": new_params,
            "return_annotation": signature.return_annotation,
        }
    )


def service_method(
    name: str | None = None,
    path: str | None = None,
    roles: list[ServiceRole] | None = None,
    autosplat: list[str] | None = None,
    warning: APIEndpointWarning | None = None,
) -> Callable:
    if roles is None or len(roles) == 0:
        # TODO: this is dangerous, we probably want to be more conservative
        roles = DATA_OWNER_ROLE_LEVEL

    def wrapper(func: Any) -> Callable:
        func_name = func.__name__
        class_name = func.__qualname__.split(".")[-2]
        _path = class_name + "." + func_name
        signature = inspect.signature(func)
        signature = signature_remove_self(signature)
        signature = signature_remove_context(signature)

        input_signature = deepcopy(signature)

        @functools.wraps(func)
        def _decorator(self: Any, *args: Any, **kwargs: Any) -> Callable:
            communication_protocol = kwargs.pop("communication_protocol", None)

            if communication_protocol:
                args, kwargs = migrate_args_and_kwargs(
                    args=args, kwargs=kwargs, to_latest_protocol=True
                )
            if autosplat is not None and len(autosplat) > 0:
                args, kwargs = reconstruct_args_kwargs(
                    signature=input_signature,
                    autosplat=autosplat,
                    args=args,
                    kwargs=kwargs,
                )
            result = func(self, *args, **kwargs)
            if communication_protocol:
                result, _ = migrate_args_and_kwargs(
                    args=(result,),
                    kwargs={},
                    to_protocol=communication_protocol,
                )
                result = result[0]
            context = kwargs.get("context", None)
            context = args[0] if context is None else context
            attrs_to_attach = {
                "syft_server_location": context.server.id,
                "syft_client_verify_key": context.credentials,
            }
            attach_attribute_to_syft_object(result=result, attr_dict=attrs_to_attach)
            return result

        if autosplat is not None and len(autosplat) > 0:
            signature = expand_signature(signature=input_signature, autosplat=autosplat)

        config = ServiceConfig(
            public_path=_path if path is None else path,
            private_path=_path,
            public_name=("public_" + func_name) if name is None else name,
            method_name=func_name,
            doc_string=func.__doc__,
            signature=signature,
            roles=roles,
            permissions=["Guest"],
            warning=warning,
        )
        ServiceConfigRegistry.register(config)

        _decorator.__name__ = func.__name__
        _decorator.__qualname__ = func.__qualname__
        return _decorator

    return wrapper


class SyftServiceRegistry:
    __service_registry__: dict[str, Callable] = {}

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if hasattr(cls, "__canonical_name__") and hasattr(cls, "__version__"):
            mapping_string = f"{cls.__canonical_name__}_{cls.__version__}"
            cls.__object_version_registry__[mapping_string] = cls

    @classmethod
    def versioned_class(cls, name: str, version: int) -> type[SyftObject] | None:
        mapping_string = f"{name}_{version}"
        if mapping_string not in cls.__object_version_registry__:
            return None
        return cls.__object_version_registry__[mapping_string]

    @classmethod
    def add_transform(
        cls,
        klass_from: str,
        version_from: int,
        klass_to: str,
        version_to: int,
        method: Callable,
    ) -> None:
        mapping_string = f"{klass_from}_{version_from}_x_{klass_to}_{version_to}"
        cls.__object_transform_registry__[mapping_string] = method

    @classmethod
    def get_transform(
        cls, type_from: type[SyftObject], type_to: type[SyftObject]
    ) -> Callable:
        klass_from = type_from.__canonical_name__
        version_from = type_from.__version__
        klass_to = type_to.__canonical_name__
        version_to = type_to.__version__
        mapping_string = f"{klass_from}_{version_from}_x_{klass_to}_{version_to}"
        return cls.__object_transform_registry__[mapping_string]


def from_api_or_context(
    func_or_path: str,
    syft_server_location: UID | None = None,
    syft_client_verify_key: SyftVerifyKey | None = None,
) -> APIModule | SyftError | partial | None:
    # relative
    from ..client.api import APIRegistry
    from ..server.server import AuthServerContextRegistry

    if callable(func_or_path):
        func_or_path = func_or_path.__qualname__

    if not (syft_server_location and syft_client_verify_key):
        return None

    api = APIRegistry.api_for(
        server_uid=syft_server_location,
        user_verify_key=syft_client_verify_key,
    )
    if api is not None:
        service_method = api.services
        for path in func_or_path.split("."):
            service_method = getattr(service_method, path)
        return service_method

    server_context = AuthServerContextRegistry.auth_context_for_user(
        server_uid=syft_server_location,
        user_verify_key=syft_client_verify_key,
    )
    if server_context is not None and server_context.server is not None:
        user_config_registry = UserServiceConfigRegistry.from_role(
            server_context.role,
        )
        if func_or_path not in user_config_registry:
            if ServiceConfigRegistry.path_exists(func_or_path):
                return SyftError(
                    message=f"As a `{server_context.role}` you have has no access to: {func_or_path}"
                )
            else:
                return SyftError(
                    message=f"API call not in registered services: {func_or_path}"
                )

        _private_api_path = user_config_registry.private_path_for(func_or_path)
        service_method = server_context.server.get_service_method(
            _private_api_path,
        )
        return partial(service_method, server_context)
    else:
        logger.error("Could not get method from api or context")
        return None



# stdlib

# third party
from result import Result

# relative
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from ...store.document_store import QueryKeys
from ...util.telemetry import instrument
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import SERVICE_TO_TYPES
from ..service import TYPE_TO_SERVICE
from .data_subject_member import ChildPartitionKey
from .data_subject_member import DataSubjectMemberRelationship
from .data_subject_member import ParentPartitionKey


@instrument
@serializable(canonical_name="DataSubjectMemberStash", version=1)
class DataSubjectMemberStash(BaseUIDStoreStash):
    object_type = DataSubjectMemberRelationship
    settings: PartitionSettings = PartitionSettings(
        name=DataSubjectMemberRelationship.__canonical_name__,
        object_type=DataSubjectMemberRelationship,
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)

    def get_all_for_parent(
        self, credentials: SyftVerifyKey, name: str
    ) -> Result[DataSubjectMemberRelationship | None, str]:
        qks = QueryKeys(qks=[ParentPartitionKey.with_obj(name)])
        return self.query_all(credentials=credentials, qks=qks)

    def get_all_for_child(
        self, credentials: SyftVerifyKey, name: str
    ) -> Result[DataSubjectMemberRelationship | None, str]:
        qks = QueryKeys(qks=[ChildPartitionKey.with_obj(name)])
        return self.query_all(credentials=credentials, qks=qks)


@instrument
@serializable(canonical_name="DataSubjectMemberService", version=1)
class DataSubjectMemberService(AbstractService):
    store: DocumentStore
    stash: DataSubjectMemberStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = DataSubjectMemberStash(store=store)

    def add(
        self, context: AuthedServiceContext, parent: str, child: str
    ) -> SyftSuccess | SyftError:
        """Register relationship between data subject and it's member."""
        relation = DataSubjectMemberRelationship(parent=parent, child=child)
        result = self.stash.set(context.credentials, relation, ignore_duplicates=True)
        if result.is_err():
            return SyftError(result.err())
        return SyftSuccess(message=f"Relationship added for: {parent} -> {child}")

    def get_relatives(
        self, context: AuthedServiceContext, data_subject_name: str
    ) -> list[str] | SyftError:
        """Get all Members for given data subject"""
        result = self.stash.get_all_for_parent(
            context.credentials, name=data_subject_name
        )
        if result.is_ok():
            data_subject_members = result.ok()
            return data_subject_members
        return SyftError(message=result.err())


TYPE_TO_SERVICE[DataSubjectMemberRelationship] = DataSubjectMemberService
SERVICE_TO_TYPES[DataSubjectMemberService].update({DataSubjectMemberRelationship})



# stdlib
from typing import cast

# third party
import numpy as np
import pyarrow as pa

# relative
from ..util.experimental_flags import ApacheArrowCompression
from ..util.experimental_flags import flags
from .deserialize import _deserialize
from .serialize import _serialize


def arrow_serialize(obj: np.ndarray) -> bytes:
    # inner function to make sure variables go out of scope after this
    def inner(obj: np.ndarray) -> tuple:
        original_dtype = obj.dtype
        apache_arrow = pa.Tensor.from_numpy(obj=obj)
        sink = pa.BufferOutputStream()
        pa.ipc.write_tensor(apache_arrow, sink)
        buffer = sink.getvalue()
        if flags.APACHE_ARROW_COMPRESSION is ApacheArrowCompression.NONE:
            numpy_bytes = buffer.to_pybytes()
        else:
            numpy_bytes = pa.compress(
                buffer, asbytes=True, codec=flags.APACHE_ARROW_COMPRESSION.value
            )
        dtype = original_dtype.name
        return (numpy_bytes, buffer.size, dtype)

    m = inner(obj)
    return cast(bytes, _serialize(m, to_bytes=True))


def arrow_deserialize(
    numpy_bytes: bytes, decompressed_size: int, dtype: str
) -> np.ndarray:
    original_dtype = np.dtype(dtype)
    if flags.APACHE_ARROW_COMPRESSION is ApacheArrowCompression.NONE:
        reader = pa.BufferReader(numpy_bytes)
        numpy_bytes = reader.read_buffer()
    else:
        numpy_bytes = pa.decompress(
            numpy_bytes,
            decompressed_size=decompressed_size,
            codec=flags.APACHE_ARROW_COMPRESSION.value,
        )

    result = pa.ipc.read_tensor(numpy_bytes)
    np_array = result.to_numpy()
    np_array.setflags(write=True)
    return np_array.astype(original_dtype)


def numpyutf8toarray(input_index: np.ndarray) -> np.ndarray:
    """Decodes utf-8 encoded numpy array to string numpy array.

    Args:
        input_index (np.ndarray): utf-8 encoded array

    Returns:
        np.ndarray: decoded NumpyArray.
    """
    shape_length = int(input_index[-1])
    shape = tuple(input_index[-(shape_length + 1) : -1])  # noqa
    string_index = input_index[: -(shape_length + 1)]
    index_length = int(string_index[-1])
    index_array = string_index[-(index_length + 1) : -1]  # noqa
    string_array: np.ndarray = string_index[: -(index_length + 1)]
    output_bytes: bytes = string_array.astype(np.uint8).tobytes()
    output_list = []
    last_offset = 0
    for offset in index_array:
        chars = output_bytes[last_offset:offset]
        final_string = chars.decode("utf-8")
        last_offset = offset
        output_list.append(final_string)
    return np.array(output_list).reshape(shape)


def arraytonumpyutf8(string_list: str | np.ndarray) -> bytes:
    """Encodes string Numpyarray  to utf-8 encoded numpy array.

    Args:
        string_list (np.ndarray): NumpyArray to be encoded

    Returns:
        bytes: serialized utf-8 encoded int Numpy array
    """
    array_shape = np.array(string_list).shape
    string_list = np.array(string_list).flatten()
    bytes_list = []
    indexes = []
    offset = 0

    for item in string_list:
        name_bytes = item.encode("utf-8")
        offset += len(name_bytes)
        indexes.append(offset)
        bytes_list.append(name_bytes)

    np_bytes = np.frombuffer(b"".join(bytes_list), dtype=np.uint8)
    np_bytes = np_bytes.astype(np.uint64)
    np_indexes = np.array(indexes, dtype=np.uint64)
    index_length = np.array([len(np_indexes)], dtype=np.uint64)
    shape = np.array(array_shape, dtype=np.uint64)
    shape_length = np.array([len(shape)], dtype=np.uint64)
    output_array = np.concatenate(
        [np_bytes, np_indexes, index_length, shape, shape_length]
    )

    return cast(bytes, _serialize(output_array, to_bytes=True))


def numpy_serialize(obj: np.ndarray) -> bytes:
    if obj.dtype.type != np.str_:
        return arrow_serialize(obj)
    else:
        return arraytonumpyutf8(obj)


def numpy_deserialize(buf: bytes) -> np.ndarray:
    deser = _deserialize(buf, from_bytes=True)
    if isinstance(deser, tuple):
        return arrow_deserialize(*deser)
    elif isinstance(deser, np.ndarray):
        return numpyutf8toarray(deser)
    else:
        raise ValueError(f"Invalid type:{type(deser)} for numpy deserialization")



# stdlib

# third party
from typing_extensions import Self

# relative
from ..service.context import ServerServiceContext
from ..service.response import SyftError
from ..service.user.user_roles import ServiceRole


class PySyftException(Exception):
    """Base class for all PySyft exceptions."""

    def __init__(self, message: str, roles: list[ServiceRole] | None = None):
        super().__init__(message)
        self.message = message
        self.roles = roles if roles else [ServiceRole.ADMIN]

    def raise_with_context(self, context: ServerServiceContext) -> Self:
        self.context = context
        return self

    def handle(self) -> SyftError:
        # if self.context and self.context.role in self.roles:
        return SyftError(message=self.message)
        # else:
        #     return SyftError(message="Access denied to exception message.")



# stdlib
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field
import typing
from typing import Any
from typing import TYPE_CHECKING

# relative
from ..serde.serializable import serializable
from ..service.action.action_service import ActionService
from ..service.action.action_store import ActionStore
from ..service.api.api_service import APIService
from ..service.attestation.attestation_service import AttestationService
from ..service.blob_storage.service import BlobStorageService
from ..service.code.status_service import UserCodeStatusService
from ..service.code.user_code_service import UserCodeService
from ..service.code_history.code_history_service import CodeHistoryService
from ..service.data_subject.data_subject_member_service import DataSubjectMemberService
from ..service.data_subject.data_subject_service import DataSubjectService
from ..service.dataset.dataset_service import DatasetService
from ..service.enclave.enclave_service import EnclaveService
from ..service.job.job_service import JobService
from ..service.log.log_service import LogService
from ..service.metadata.metadata_service import MetadataService
from ..service.migration.migration_service import MigrationService
from ..service.network.network_service import NetworkService
from ..service.notification.notification_service import NotificationService
from ..service.notifier.notifier_service import NotifierService
from ..service.output.output_service import OutputService
from ..service.policy.policy_service import PolicyService
from ..service.project.project_service import ProjectService
from ..service.queue.queue_service import QueueService
from ..service.request.request_service import RequestService
from ..service.service import AbstractService
from ..service.settings.settings_service import SettingsService
from ..service.sync.sync_service import SyncService
from ..service.user.user_service import UserService
from ..service.worker.image_registry_service import SyftImageRegistryService
from ..service.worker.worker_image_service import SyftWorkerImageService
from ..service.worker.worker_pool_service import SyftWorkerPoolService
from ..service.worker.worker_service import WorkerService

if TYPE_CHECKING:
    # relative
    from .server import Server


@serializable(canonical_name="ServiceRegistry", version=1)
@dataclass
class ServiceRegistry:
    action: ActionService
    user: UserService
    attestation: AttestationService
    worker: WorkerService
    settings: SettingsService
    dataset: DatasetService
    user_code: UserCodeService
    log: LogService
    request: RequestService
    queue: QueueService
    job: JobService
    api: APIService
    data_subject: DataSubjectService
    network: NetworkService
    policy: PolicyService
    notifier: NotifierService
    notification: NotificationService
    data_subject_member: DataSubjectMemberService
    project: ProjectService
    enclave: EnclaveService
    code_history: CodeHistoryService
    metadata: MetadataService
    blob_storage: BlobStorageService
    migration: MigrationService
    syft_worker_image: SyftWorkerImageService
    syft_worker_pool: SyftWorkerPoolService
    syft_image_registry: SyftImageRegistryService
    sync: SyncService
    output: OutputService
    user_code_status: UserCodeStatusService

    services: list[AbstractService] = field(default_factory=list, init=False)
    service_path_map: dict[str, AbstractService] = field(
        default_factory=dict, init=False
    )

    @classmethod
    def for_server(cls, server: "Server") -> "ServiceRegistry":
        return cls(**cls._construct_services(server))

    def __post_init__(self) -> None:
        for name, service_cls in self.get_service_classes().items():
            service = getattr(self, name)
            self.services.append(service)
            self.service_path_map[service_cls.__name__.lower()] = service

    @classmethod
    def get_service_classes(
        cls,
    ) -> dict[str, type[AbstractService]]:
        return {
            name: cls
            for name, cls in typing.get_type_hints(cls).items()
            if issubclass(cls, AbstractService)
        }

    @classmethod
    def _construct_services(cls, server: "Server") -> dict[str, AbstractService]:
        service_dict = {}
        for field_name, service_cls in cls.get_service_classes().items():
            svc_kwargs: dict[str, Any] = {}
            if issubclass(service_cls.store_type, ActionStore):
                svc_kwargs["store"] = server.action_store
            else:
                svc_kwargs["store"] = server.document_store

            service = service_cls(**svc_kwargs)
            service_dict[field_name] = service
        return service_dict

    def get_service(self, path_or_func: str | Callable) -> AbstractService:
        if callable(path_or_func):
            path_or_func = path_or_func.__qualname__
        return self._get_service_from_path(path_or_func)

    def _get_service_from_path(self, path: str) -> AbstractService:
        try:
            path_list = path.split(".")
            if len(path_list) > 1:
                _ = path_list.pop()
            service_name = path_list.pop()
            return self.service_path_map[service_name.lower()]
        except KeyError:
            raise ValueError(f"Service {path} not found.")



# stdlib
import logging
from typing import Any

# relative
from .action_data_empty import ActionDataEmpty

logger = logging.getLogger(__name__)

action_types: dict = {}


def action_type_for_type(obj_or_type: Any) -> type:
    """Convert standard type to Syft types

    Parameters:
        obj_or_type: Union[object, type]
            Can be an object or a class
    """
    if isinstance(obj_or_type, ActionDataEmpty):
        obj_or_type = obj_or_type.syft_internal_type
    if type(obj_or_type) != type:
        obj_or_type = type(obj_or_type)

    if obj_or_type not in action_types:
        logger.debug(
            f"WARNING: No Type for {obj_or_type}, returning {action_types[Any]}"
        )

    return action_types.get(obj_or_type, action_types[Any])


def action_type_for_object(obj: Any) -> type:
    """Convert standard type to Syft types

    Parameters:
        obj_or_type: Union[object, type]
            Can be an object or a class
    """
    _type = type(obj)

    if _type not in action_types:
        logger.debug(f"WARNING: No Type for {_type}, returning {action_types[Any]}")
        return action_types[Any]

    return action_types[_type]



# stdlib
import logging
from multiprocessing import Process
import threading
from threading import Thread
import time
from typing import Any

# third party
import psutil
from result import Err
from result import Ok

# relative
from ...serde.deserialize import _deserialize as deserialize
from ...serde.serializable import serializable
from ...server.credentials import SyftVerifyKey
from ...server.worker_settings import WorkerSettings
from ...service.context import AuthedServiceContext
from ...store.document_store import BaseStash
from ...types.datetime import DateTime
from ...types.uid import UID
from ..job.job_stash import Job
from ..job.job_stash import JobStatus
from ..response import SyftError
from ..response import SyftSuccess
from ..worker.worker_stash import WorkerStash
from .base_queue import AbstractMessageHandler
from .base_queue import BaseQueueManager
from .base_queue import QueueConfig
from .base_queue import QueueConsumer
from .base_queue import QueueProducer
from .queue_stash import QueueItem
from .queue_stash import Status

logger = logging.getLogger(__name__)


class MonitorThread(threading.Thread):
    def __init__(
        self,
        queue_item: QueueItem,
        worker: Any,  # should be of type Worker(Server), but get circular import error
        credentials: SyftVerifyKey,
        interval: int = 5,
    ) -> None:
        super().__init__()
        self.interval = interval
        self.stop_requested = threading.Event()
        self.credentials = credentials
        self.worker = worker
        self.queue_item = queue_item

    def run(self) -> None:
        while not self.stop_requested.is_set():
            self.monitor()
            time.sleep(self.interval)

    def monitor(self) -> None:
        # Implement the monitoring logic here
        job = self.worker.job_stash.get_by_uid(
            self.credentials, self.queue_item.job_id
        ).ok()
        if job and job.status == JobStatus.TERMINATING:
            self.terminate(job)
            for subjob in job.subjobs:
                self.terminate(subjob)

            self.queue_item.status = Status.INTERRUPTED
            self.queue_item.resolved = True
            self.worker.queue_stash.set_result(self.credentials, self.queue_item)
            # How about subjobs of subjobs?

    def stop(self) -> None:
        self.stop_requested.set()

    def terminate(self, job: Job) -> None:
        job.resolved = True
        job.status = JobStatus.INTERRUPTED
        self.worker.job_stash.set_result(self.credentials, job)
        try:
            psutil.Process(job.job_pid).terminate()
        except psutil.Error as e:
            logger.warning(f"Failed to terminate job {job.id}: {e}")


@serializable(canonical_name="QueueManager", version=1)
class QueueManager(BaseQueueManager):
    config: QueueConfig

    def post_init(self) -> None:
        self.client_config = self.config.client_config
        self._client = self.config.client_type(self.client_config)

    def close(self) -> SyftError | SyftSuccess:
        return self._client.close()

    def create_consumer(
        self,
        message_handler: type[AbstractMessageHandler],
        service_name: str,
        worker_stash: WorkerStash | None = None,
        address: str | None = None,
        syft_worker_id: UID | None = None,
    ) -> QueueConsumer:
        consumer = self._client.add_consumer(
            message_handler=message_handler,
            queue_name=message_handler.queue_name,
            address=address,
            service_name=service_name,
            worker_stash=worker_stash,
            syft_worker_id=syft_worker_id,
        )
        return consumer

    def create_producer(
        self,
        queue_name: str,
        queue_stash: type[BaseStash],
        context: AuthedServiceContext,
        worker_stash: WorkerStash,
    ) -> QueueProducer:
        return self._client.add_producer(
            queue_name=queue_name,
            queue_stash=queue_stash,
            context=context,
            worker_stash=worker_stash,
        )

    def send(
        self,
        message: bytes,
        queue_name: str,
    ) -> SyftSuccess | SyftError:
        return self._client.send_message(
            message=message,
            queue_name=queue_name,
        )

    @property
    def producers(self) -> Any:
        return self._client.producers

    @property
    def consumers(self) -> Any:
        return self._client.consumers


def handle_message_multiprocessing(
    worker_settings: WorkerSettings,
    queue_item: QueueItem,
    credentials: SyftVerifyKey,
) -> None:
    # this is a temp hack to prevent some multithreading issues
    time.sleep(0.5)
    queue_config = worker_settings.queue_config
    if queue_config is None:
        raise ValueError(f"{worker_settings} has no queue configurations!")
    queue_config.client_config.create_producer = False
    queue_config.client_config.n_consumers = 0

    # relative
    from ...server.server import Server

    worker = Server(
        id=worker_settings.id,
        name=worker_settings.name,
        signing_key=worker_settings.signing_key,
        document_store_config=worker_settings.document_store_config,
        action_store_config=worker_settings.action_store_config,
        blob_storage_config=worker_settings.blob_store_config,
        server_side_type=worker_settings.server_side_type,
        queue_config=queue_config,
        is_subprocess=True,
        migrate=False,
    )

    # Set monitor thread for this job.
    monitor_thread = MonitorThread(queue_item, worker, credentials)
    monitor_thread.start()

    if queue_item.service == "user":
        queue_item.service = "userservice"

    try:
        call_method = getattr(worker.get_service(queue_item.service), queue_item.method)
        role = worker.get_role_for_credentials(credentials=credentials)

        context = AuthedServiceContext(
            server=worker,
            credentials=credentials,
            role=role,
            job_id=queue_item.job_id,
            has_execute_permissions=queue_item.has_execute_permissions,
        )

        # relative
        from ...server.server import AuthServerContextRegistry

        AuthServerContextRegistry.set_server_context(
            server_uid=worker.id,
            context=context,
            user_verify_key=credentials,
        )

        result: Any = call_method(context, *queue_item.args, **queue_item.kwargs)
        status = Status.COMPLETED
        job_status = JobStatus.COMPLETED

        if isinstance(result, Ok):
            result = result.ok()
            if hasattr(result, "syft_action_data") and isinstance(
                result.syft_action_data, Err
            ):
                status = Status.ERRORED
                job_status = JobStatus.ERRORED
        elif isinstance(result, SyftError) or isinstance(result, Err):
            status = Status.ERRORED
            job_status = JobStatus.ERRORED

        else:
            raise Exception(f"Unknown result type: {type(result)}")
    except Exception as e:
        status = Status.ERRORED
        job_status = JobStatus.ERRORED
        logger.error("Unhandled error in handle_message_multiprocessing", exc_info=e)

    queue_item.result = result
    queue_item.resolved = True
    queue_item.status = status

    # get new job item to get latest iter status
    job_item = worker.job_stash.get_by_uid(credentials, queue_item.job_id).ok()
    if job_item is None:
        raise Exception(f"Job {queue_item.job_id} not found!")

    job_item.server_uid = worker.id
    job_item.result = result
    job_item.resolved = True
    job_item.status = job_status

    worker.queue_stash.set_result(credentials, queue_item)
    worker.job_stash.set_result(credentials, job_item)

    # Finish monitor thread
    monitor_thread.stop()


@serializable(canonical_name="APICallMessageHandler", version=1)
class APICallMessageHandler(AbstractMessageHandler):
    queue_name = "api_call"

    @staticmethod
    def handle_message(message: bytes, syft_worker_id: UID) -> None:
        # relative
        from ...server.server import Server

        queue_item = deserialize(message, from_bytes=True)
        worker_settings = queue_item.worker_settings

        queue_config = worker_settings.queue_config
        queue_config.client_config.create_producer = False
        queue_config.client_config.n_consumers = 0

        worker = Server(
            id=worker_settings.id,
            name=worker_settings.name,
            signing_key=worker_settings.signing_key,
            document_store_config=worker_settings.document_store_config,
            action_store_config=worker_settings.action_store_config,
            blob_storage_config=worker_settings.blob_store_config,
            server_side_type=worker_settings.server_side_type,
            queue_config=queue_config,
            is_subprocess=True,
            migrate=False,
        )

        # otherwise it reads it from env, resulting in the wrong credentials
        worker.id = worker_settings.id
        worker.signing_key = worker_settings.signing_key

        credentials = queue_item.syft_client_verify_key
        res = worker.job_stash.get_by_uid(credentials, queue_item.job_id)
        if res.is_err():
            logger.warning(res.err())
            raise Exception(res.value)
        job_item: Job = res.ok()

        queue_item.status = Status.PROCESSING
        queue_item.server_uid = worker.id

        job_item.status = JobStatus.PROCESSING
        job_item.server_uid = worker.id  # type: ignore[assignment]
        job_item.updated_at = DateTime.now()

        if syft_worker_id is not None:
            job_item.job_worker_id = syft_worker_id

        queue_result = worker.queue_stash.set_result(credentials, queue_item)
        if isinstance(queue_result, SyftError):
            raise Exception(f"{queue_result.err()}")

        job_result = worker.job_stash.set_result(credentials, job_item)
        if isinstance(job_result, SyftError):
            raise Exception(f"{job_result.err()}")

        logger.info(
            f"Handling queue item: id={queue_item.id}, method={queue_item.method} "
            f"args={queue_item.args}, kwargs={queue_item.kwargs} "
            f"service={queue_item.service}, as_thread={queue_config.thread_workers}"
        )

        if queue_config.thread_workers:
            thread = Thread(
                target=handle_message_multiprocessing,
                args=(worker_settings, queue_item, credentials),
            )
            thread.start()
            thread.join()
        else:
            # if psutil.pid_exists(job_item.job_pid):
            #     psutil.Process(job_item.job_pid).terminate()
            process = Process(
                target=handle_message_multiprocessing,
                args=(worker_settings, queue_item, credentials),
            )
            process.start()
            job_item.job_pid = process.pid
            worker.job_stash.set_result(credentials, job_item)
            process.join()



# stdlib
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any

# third party
from pydantic import BaseModel

__all__ = [
    "BuilderBase",
    "ImageBuildResult",
    "ImagePushResult",
    "BUILD_IMAGE_TIMEOUT_SEC",
    "PUSH_IMAGE_TIMEOUT_SEC",
]


BUILD_IMAGE_TIMEOUT_SEC = 30 * 60
PUSH_IMAGE_TIMEOUT_SEC = 10 * 60


class ImageBuildResult(BaseModel):
    image_hash: str
    logs: str


class ImagePushResult(BaseModel):
    logs: str
    exit_code: int


class BuilderBase(ABC):
    @abstractmethod
    def build_image(
        self,
        tag: str,
        dockerfile: str | None = None,
        dockerfile_path: Path | None = None,
        buildargs: dict | None = None,
        **kwargs: Any,
    ) -> ImageBuildResult:
        pass

    @abstractmethod
    def push_image(
        self,
        tag: str,
        registry_url: str,
        username: str | None = None,
        password: str | None = None,
        **kwargs: Any,
    ) -> ImagePushResult:
        pass



# stdlib
from collections.abc import Callable
from collections.abc import Sequence
import importlib
import inspect
from inspect import Signature
from inspect import _signature_fromstr
from types import BuiltinFunctionType
from typing import Any

# third party
import numpy
from typing_extensions import Self

# relative
from .lib_permissions import ALL_EXECUTE
from .lib_permissions import CMPPermission
from .lib_permissions import NONE_EXECUTE
from .signature import get_signature

LIB_IGNORE_ATTRIBUTES = {
    "os",
    "__abstractmethods__",
    "__base__",
    " __bases__",
    "__class__",
}


def import_from_path(path: str) -> type:
    if "." in path:
        top_level_module, attr_path = path.split(".", 1)
    else:
        top_level_module = path
        attr_path = ""

    res = importlib.import_module(top_level_module)
    path_parts = [x for x in attr_path.split(".") if x != ""]
    for attr in path_parts:
        res = getattr(res, attr)
    return res  # type: ignore


class CMPBase:
    """cmp: cascading module permissions"""

    def __init__(
        self,
        path: str,
        children: list | dict | None = None,
        permissions: CMPPermission | None = None,
        obj: Any | None = None,
        absolute_path: str | None = None,
        text_signature: str | None = None,
    ):
        self.permissions: CMPPermission | None = permissions
        self.path: str = path
        self.obj: Any | None = obj if obj is not None else None
        self.absolute_path = absolute_path
        self.signature: Signature | None = None

        self.children: dict[str, CMPBase] = {}
        if isinstance(children, list):
            self.children = {f"{c.path}": c for c in children}
        elif isinstance(children, dict):
            self.children = children

        for c in self.children.values():
            if c.absolute_path is None:
                c.absolute_path = f"{path}.{c.path}"

        if text_signature is not None:
            self.signature = _signature_fromstr(
                inspect.Signature, obj, text_signature, True
            )

        self.is_built = False

    def set_signature(self) -> None:
        pass

    def build(self) -> None:
        if self.obj is None:
            self.obj = import_from_path(self.absolute_path)

        if self.signature is None:
            self.set_signature()

        child_paths = set(self.children.keys())

        for attr_name in getattr(self.obj, "__dict__", {}).keys():
            if attr_name not in LIB_IGNORE_ATTRIBUTES:
                if attr_name in child_paths:
                    child = self.children[attr_name]
                else:
                    try:
                        attr = getattr(self.obj, attr_name)
                    except Exception:  # nosec
                        continue
                    child = self.init_child(  # type: ignore
                        self.obj,
                        f"{self.path}.{attr_name}",
                        attr,
                        f"{self.absolute_path}.{attr_name}",
                    )
                if child is not None:
                    child.build()
                    self.children[attr_name] = child

    def __getattr__(self, __name: str) -> Any:
        if __name in self.children:
            return self.children[__name]
        else:
            raise ValueError(f"property {__name} not defined")

    def init_child(
        self,
        parent_obj: type | object,
        child_path: str,
        child_obj: type | object,
        absolute_path: str,
    ) -> Self | None:
        """Get the child of parent as a CMPBase object

        Args:
            parent_obj (_type_): parent object
            child_path (_type_): _description_
            child_obj (_type_): _description_

        Returns:
            _type_: _description_
        """
        parent_is_parent_module = CMPBase.parent_is_parent_module(parent_obj, child_obj)
        if CMPBase.isfunction(child_obj) and parent_is_parent_module:
            return CMPFunction(
                child_path,
                permissions=self.permissions,
                obj=child_obj,
                absolute_path=absolute_path,
            )  # type: ignore
        elif inspect.ismodule(child_obj) and CMPBase.is_submodule(
            parent_obj, child_obj
        ):
            ## TODO, we could register modules and functions in 2 ways:
            # A) as numpy.float32 (what we are doing now)
            # B) as numpy.core.float32 (currently not supported)
            # only allow submodules

            return CMPModule(
                child_path,
                permissions=self.permissions,
                obj=child_obj,
                absolute_path=absolute_path,
            )  # type: ignore
        elif inspect.isclass(child_obj) and parent_is_parent_module:
            return CMPClass(
                child_path,
                permissions=self.permissions,
                obj=child_obj,
                absolute_path=absolute_path,
            )  # type: ignore
        else:
            return None

    @staticmethod
    def is_submodule(parent: type, child: type) -> bool:
        try:
            if "." not in child.__package__:
                return False
            else:
                child_parent_module = child.__package__.rsplit(".", 1)[0]
                if parent.__package__ == child_parent_module:
                    return True
                else:
                    return False
        except Exception:  # nosec
            pass
        return False

    @staticmethod
    def parent_is_parent_module(parent_obj: Any, child_obj: Any) -> str | None:
        try:
            if hasattr(child_obj, "__module__"):
                return child_obj.__module__ == parent_obj.__name__
            else:
                # TODO: this is a fix for for instance numpy ufuncs
                return child_obj.__class__.__module__ == parent_obj.__name__
        except Exception:  # nosec
            pass
        return None

    def flatten(self) -> list[Self]:
        res = [self]
        for c in self.children.values():
            res += c.flatten()
        return res

    @staticmethod
    def isfunction(obj: Callable) -> bool:
        return (
            inspect.isfunction(obj)
            or type(obj) == numpy.ufunc
            or isinstance(obj, BuiltinFunctionType)
        )

    def __repr__(
        self, indent: int = 0, is_last: bool = False, parent_path: str = ""
    ) -> str:
        """Visualize the tree, e.g.:
        ├───numpy (ALL_EXECUTE)
        │    ├───ModuleDeprecationWarning (ALL_EXECUTE)
        │    ├───VisibleDeprecationWarning (ALL_EXECUTE)
        │    ├───_CopyMode (ALL_EXECUTE)
        │    ├───compat (ALL_EXECUTE)
        │    ├───core (ALL_EXECUTE)
        │    │    ├───_ufunc_reconstruct (ALL_EXECUTE)
        │    │    ├───_DType_reconstruct (ALL_EXECUTE)
        │    │    └───__getattr__ (ALL_EXECUTE)
        │    ├───char (ALL_EXECUTE)
        │    │    ├───_use_unicode (ALL_EXECUTE)
        │    │    ├───_to_string_or_unicode_array (ALL_EXECUTE)
        │    │    ├───_clean_args (ALL_EXECUTE)

        Args:
            indent (int, optional): indentation level. Defaults to 0.
            is_last (bool, optional): is last item of collection. Defaults to False.
            parent_path (str, optional): path of the parent obj. Defaults to "".

        Returns:
            str: representation of the CMP
        """
        last_idx, c_indent = len(self.children) - 1, indent + 1
        children_string = "".join(
            [
                c.__repr__(c_indent, is_last=i == last_idx, parent_path=self.path)  # type: ignore
                for i, c in enumerate(
                    sorted(
                        self.children.values(),
                        key=lambda x: x.permissions.permission_string,  # type: ignore
                    )
                )
            ]
        )
        tree_prefix = "└───" if is_last else "├───"
        indent_str = "│    " * indent + tree_prefix
        if parent_path != "":
            path = self.path.replace(f"{parent_path}.", "")
        else:
            path = self.path
        return f"{indent_str}{path} ({self.permissions})\n{children_string}"


class CMPModule(CMPBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class CMPFunction(CMPBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.set_signature()

    @property
    def name(self) -> str:
        return self.obj.__name__  # type: ignore

    def set_signature(self) -> None:
        try:
            self.signature = get_signature(self.obj)
        except Exception:  # nosec
            pass


class CMPClass(CMPBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.set_signature()

    @property
    def name(self) -> str:
        # possibly change to
        # func_name = path.split(".")[-1]
        return self.obj.__name__  # type: ignore

    def set_signature(self) -> None:
        try:
            self.signature = get_signature(self.obj)
        except Exception:  # nosec
            try:
                self.signature = get_signature(self.obj.__init__)  # type: ignore
            except Exception:  # nosec
                pass


class CMPMethod(CMPBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class CMPProperty(CMPBase):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)


class CMPTree:
    """root node of the Tree(s), with one child per library"""

    def __init__(self, children: list[CMPModule]):
        self.children = {c.path: c for c in children}

    def build(self) -> Self:
        for c in self.children.values():
            c.absolute_path = c.path
            c.build()
        return self

    def flatten(self) -> Sequence[CMPBase]:
        res = []
        for c in self.children.values():
            res += c.flatten()
        return res

    def __getattr__(self, _name: str) -> Any:
        if _name in self.children:
            return self.children[_name]
        else:
            raise ValueError(f"property {_name} does not exist")

    def __repr__(self) -> str:
        return "\n".join([c.__repr__() for c in self.children.values()])


action_execute_registry_libs = CMPTree(
    children=[
        CMPModule(
            "numpy",
            permissions=ALL_EXECUTE,
            children=[
                CMPFunction(
                    "concatenate",
                    permissions=ALL_EXECUTE,
                    text_signature="concatenate(a1,a2, *args,axis=0,out=None,dtype=None,casting='same_kind')",
                ),
                CMPFunction("source", permissions=NONE_EXECUTE),
                CMPFunction("fromfile", permissions=NONE_EXECUTE),
                CMPFunction(
                    "set_numeric_ops",
                    permissions=ALL_EXECUTE,
                    text_signature="set_numeric_ops(op1,op2, *args)",
                ),
                CMPModule("testing", permissions=NONE_EXECUTE),
            ],
        ),
    ]
).build()



# relative
from ...serde.serializable import serializable
from ...store.document_store import BaseUIDStoreStash
from ...store.document_store import DocumentStore
from ...store.document_store import PartitionSettings
from ...util.telemetry import instrument
from .log import SyftLog


@instrument
@serializable(canonical_name="LogStash", version=1)
class LogStash(BaseUIDStoreStash):
    object_type = SyftLog
    settings: PartitionSettings = PartitionSettings(
        name=SyftLog.__canonical_name__, object_type=SyftLog
    )

    def __init__(self, store: DocumentStore) -> None:
        super().__init__(store=store)



# stdlib
from typing import Any

# third party
from typing_extensions import Self

# relative
from ..serde.serializable import serializable
from ..service.context import AuthedServiceContext
from ..service.context import ChangeContext
from ..service.context import ServerServiceContext
from ..service.response import SyftError
from ..service.response import SyftSuccess
from ..types.syft_object import SYFT_OBJECT_VERSION_1
from ..types.syft_object import SyftObject
from ..types.uid import UID


@serializable()
class LinkedObject(SyftObject):
    __canonical_name__ = "LinkedObject"
    __version__ = SYFT_OBJECT_VERSION_1

    server_uid: UID
    service_type: type[Any]
    object_type: type[SyftObject]
    object_uid: UID

    _resolve_cache: SyftObject | None = None

    __exclude_sync_diff_attrs__ = ["server_uid"]

    def __str__(self) -> str:
        resolved_obj_type = (
            type(self.resolve) if self.object_type is None else self.object_type
        )
        return f"{resolved_obj_type.__name__}: {self.object_uid} @ Server {self.server_uid}"

    @property
    def resolve(self) -> SyftObject:
        # relative
        from ..client.api import APIRegistry

        api = APIRegistry.api_for(
            server_uid=self.server_uid,
            user_verify_key=self.syft_client_verify_key,
        )
        if api is None:
            raise ValueError(f"api is None. You must login to {self.server_uid}")

        resolve: SyftObject = api.services.notifications.resolve_object(self)
        self._resolve_cache = resolve
        return resolve

    def resolve_with_context(self, context: ServerServiceContext) -> Any:
        if context.server is None:
            raise ValueError(f"context {context}'s server is None")
        return context.server.get_service(self.service_type).resolve_link(
            context=context, linked_obj=self
        )

    def update_with_context(
        self, context: ServerServiceContext | ChangeContext | Any, obj: Any
    ) -> SyftSuccess | SyftError:
        if isinstance(context, AuthedServiceContext):
            credentials = context.credentials
        elif isinstance(context, ChangeContext):
            credentials = context.approving_user_credentials
        else:
            return SyftError(message="wrong context passed")
        if context.server is None:
            return SyftError(message=f"context {context}'s server is None")
        service = context.server.get_service(self.service_type)
        if hasattr(service, "stash"):
            result = service.stash.update(credentials, obj)
        else:
            return SyftError(message=f"service {service} does not have a stash")
        return result

    @classmethod
    def from_obj(
        cls,
        obj: SyftObject | type[SyftObject],
        service_type: type[Any] | None = None,
        server_uid: UID | None = None,
    ) -> Self:
        if service_type is None:
            # relative
            from ..service.action.action_object import ActionObject
            from ..service.action.action_service import ActionService
            from ..service.service import TYPE_TO_SERVICE

            if isinstance(obj, ActionObject):
                service_type = ActionService
            else:
                service_type = TYPE_TO_SERVICE[type(obj)]

        object_uid = getattr(obj, "id", None)
        if object_uid is None:
            raise Exception(f"{cls} Requires an object UID")

        if server_uid is None:
            server_uid = getattr(obj, "server_uid", None)
            if server_uid is None:
                raise Exception(f"{cls} Requires an object UID")

        return LinkedObject(
            server_uid=server_uid,
            service_type=service_type,
            object_type=type(obj),
            object_uid=object_uid,
            syft_client_verify_key=obj.syft_client_verify_key,
        )

    @classmethod
    def with_context(
        cls,
        obj: SyftObject,
        context: ServerServiceContext,
        object_uid: UID | None = None,
        service_type: type[Any] | None = None,
    ) -> Self:
        if service_type is None:
            # relative
            from ..service.service import TYPE_TO_SERVICE

            service_type = TYPE_TO_SERVICE[type(obj)]

        if object_uid is None and hasattr(obj, "id"):
            object_uid = getattr(obj, "id", None)
        if object_uid is None:
            raise Exception(f"{cls} Requires an object UID")

        if context.server is None:
            raise ValueError(f"context {context}'s server is None")
        server_uid = context.server.id

        return LinkedObject(
            server_uid=server_uid,
            service_type=service_type,
            object_type=type(obj),
            object_uid=object_uid,
        )

    @classmethod
    def from_uid(
        cls,
        object_uid: UID,
        object_type: type[SyftObject],
        service_type: type[Any],
        server_uid: UID,
    ) -> Self:
        return cls(
            server_uid=server_uid,
            service_type=service_type,
            object_type=object_type,
            object_uid=object_uid,
        )



# stdlib
import time
from typing import Any
from typing import cast

# third party
from pydantic import ValidationError
from result import Err
from result import Ok

# relative
from ...serde.serializable import serializable
from ...service.action.action_endpoint import CustomEndpointActionObject
from ...service.action.action_object import ActionObject
from ...store.document_store import DocumentStore
from ...types.uid import UID
from ...util.telemetry import instrument
from ..action.action_service import ActionService
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import TYPE_TO_SERVICE
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from ..user.user_roles import DATA_SCIENTIST_ROLE_LEVEL
from ..user.user_roles import GUEST_ROLE_LEVEL
from .api import CreateTwinAPIEndpoint
from .api import Endpoint
from .api import PrivateAPIEndpoint
from .api import PublicAPIEndpoint
from .api import TwinAPIContextView
from .api import TwinAPIEndpoint
from .api import TwinAPIEndpointView
from .api import UpdateTwinAPIEndpoint
from .api_stash import TwinAPIEndpointStash


@instrument
@serializable(canonical_name="APIService", version=1)
class APIService(AbstractService):
    store: DocumentStore
    stash: TwinAPIEndpointStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = TwinAPIEndpointStash(store=store)

    @service_method(
        path="api.add",
        name="add",
        roles=ADMIN_ROLE_LEVEL,
    )
    def set(
        self,
        context: AuthedServiceContext,
        endpoint: CreateTwinAPIEndpoint | TwinAPIEndpoint,
    ) -> SyftSuccess | SyftError:
        """Register an CustomAPIEndpoint."""
        try:
            new_endpoint = None
            if isinstance(endpoint, CreateTwinAPIEndpoint):  # type: ignore
                new_endpoint = endpoint.to(TwinAPIEndpoint)
            elif isinstance(endpoint, TwinAPIEndpoint):  # type: ignore
                new_endpoint = endpoint

            if new_endpoint is None:
                return SyftError(message="Invalid endpoint type.")  # type: ignore
        except ValueError as e:
            return SyftError(message=str(e))

        if isinstance(endpoint, CreateTwinAPIEndpoint):
            endpoint_exists = self.stash.path_exists(
                context.credentials, new_endpoint.path
            )

            if endpoint_exists.is_err():
                return SyftError(message=endpoint_exists.err())

            if endpoint_exists.is_ok() and endpoint_exists.ok():
                return SyftError(
                    message="An API endpoint already exists at the given path."
                )

        result = self.stash.upsert(context.credentials, endpoint=new_endpoint)
        if result.is_err():
            return SyftError(message=result.err())

        result = result.ok()
        action_obj = ActionObject.from_obj(
            id=new_endpoint.action_object_id,
            syft_action_data=CustomEndpointActionObject(endpoint_id=result.id),
            syft_server_location=context.server.id,
            syft_client_verify_key=context.credentials,
        )
        action_service = context.server.get_service("actionservice")
        res = action_service.set_result_to_store(
            context=context,
            result_action_object=action_obj,
            has_result_read_permission=True,
        )
        if res.is_err():
            return SyftError(message=res.err())

        return SyftSuccess(message="Endpoint successfully created.")

    @service_method(
        path="api.update",
        name="update",
        roles=ADMIN_ROLE_LEVEL,
    )
    def update(
        self,
        context: AuthedServiceContext,
        endpoint_path: str,
        mock_function: Endpoint | None = None,
        private_function: Endpoint | None = None,
        hide_mock_definition: bool | None = None,
        endpoint_timeout: int | None = None,
    ) -> SyftSuccess | SyftError:
        """Updates an specific API endpoint."""

        # if any of these are supplied e.g. truthy then keep going otherwise return
        # an error
        # TODO: change to an Update object with autosplat
        if not (
            mock_function
            or private_function
            or (hide_mock_definition is not None)
            or endpoint_timeout
        ):
            return SyftError(
                message='At least one of "mock_function", "private_function", '
                '"hide_mock_definition" or "endpoint_timeout" is required.'
            )

        endpoint_result = self.stash.get_by_path(context.credentials, endpoint_path)

        if endpoint_result.is_err():
            return SyftError(message=endpoint_result.err())

        if not endpoint_result.ok():
            return SyftError(message=f"Enpoint at path {endpoint_path} doesn't exist")

        endpoint: TwinAPIEndpoint = endpoint_result.ok()

        endpoint_timeout = (
            endpoint_timeout
            if endpoint_timeout is not None
            else endpoint.endpoint_timeout
        )

        updated_mock = (
            mock_function.to(PublicAPIEndpoint)
            if mock_function is not None
            else endpoint.mock_function
        )
        updated_private = (
            private_function.to(PrivateAPIEndpoint)
            if private_function is not None
            else endpoint.private_function
        )

        try:
            endpoint_update = UpdateTwinAPIEndpoint(
                path=endpoint_path,
                mock_function=updated_mock,
                private_function=updated_private,
                endpoint_timeout=endpoint_timeout,
            )
        except ValidationError as e:
            return SyftError(message=str(e))

        endpoint.mock_function = endpoint_update.mock_function
        endpoint.private_function = endpoint_update.private_function
        endpoint.signature = updated_mock.signature
        endpoint.endpoint_timeout = endpoint_update.endpoint_timeout

        if hide_mock_definition is not None:
            view_access = not hide_mock_definition
            endpoint.mock_function.view_access = view_access

        # save changes
        result = self.stash.upsert(context.credentials, endpoint=endpoint)
        if result.is_err():
            return SyftError(message=result.err())

        return SyftSuccess(message="Endpoint successfully updated.")

    @service_method(
        path="api.delete",
        name="delete",
        roles=ADMIN_ROLE_LEVEL,
    )
    def delete(
        self, context: AuthedServiceContext, endpoint_path: str
    ) -> SyftSuccess | SyftError:
        """Deletes an specific API endpoint."""

        result = self.stash.get_by_path(context.credentials, endpoint_path)

        if result.is_err():
            return SyftError(message=result.err())

        endpoint = result.ok()
        if not endpoint:
            return SyftError(message=f"Enpoint at path {endpoint_path} doesn't exist")

        result = self.stash.delete_by_uid(context.credentials, endpoint.id)

        if result.is_err():
            return SyftError(message=result.err())

        return SyftSuccess(message="Endpoint successfully deleted.")

    @service_method(
        path="api.view",
        name="view",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def view(
        self, context: AuthedServiceContext, path: str
    ) -> TwinAPIEndpointView | SyftError:
        """Retrieves an specific API endpoint."""
        result = self.stash.get_by_path(context.server.verify_key, path)
        if result.is_err():
            return SyftError(message=result.err())
        api_endpoint = result.ok()

        return api_endpoint.to(TwinAPIEndpointView, context=context)

    @service_method(
        path="api.get",
        name="get",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get(
        self, context: AuthedServiceContext, api_path: str
    ) -> TwinAPIEndpoint | SyftError:
        """Retrieves an specific API endpoint."""
        result = self.stash.get_by_path(context.server.verify_key, api_path)
        if result.is_err():
            return SyftError(message=result.err())
        api_endpoint = result.ok()

        return api_endpoint

    @service_method(
        path="api.set_state",
        name="set_state",
        roles=ADMIN_ROLE_LEVEL,
    )
    def set_state(
        self,
        context: AuthedServiceContext,
        api_path: str,
        state: dict,
        private: bool = False,
        mock: bool = False,
        both: bool = False,
    ) -> TwinAPIEndpoint | SyftError:
        """Sets the state of a specific API endpoint."""
        if both:
            private = True
            mock = True
        result = self.stash.get_by_path(context.server.verify_key, api_path)
        if result.is_err():
            return SyftError(message=result.err())
        api_endpoint = result.ok()

        if private and api_endpoint.private_function:
            api_endpoint.private_function.state = state
        if mock and api_endpoint.mock_function:
            api_endpoint.mock_function.state = state

        result = self.stash.upsert(context.credentials, endpoint=api_endpoint)
        if result.is_err():
            return SyftError(message=result.err())
        return SyftSuccess(message=f"APIEndpoint {api_path} state updated.")

    @service_method(
        path="api.set_settings",
        name="set_settings",
        roles=ADMIN_ROLE_LEVEL,
    )
    def set_settings(
        self,
        context: AuthedServiceContext,
        api_path: str,
        settings: dict,
        private: bool = False,
        mock: bool = False,
        both: bool = False,
    ) -> TwinAPIEndpoint | SyftError:
        """Sets the settings of a specific API endpoint."""
        if both:
            private = True
            mock = True
        result = self.stash.get_by_path(context.server.verify_key, api_path)
        if result.is_err():
            return SyftError(message=result.err())
        api_endpoint = result.ok()

        if private and api_endpoint.private_function:
            api_endpoint.private_function.settings = settings
        if mock and api_endpoint.mock_function:
            api_endpoint.mock_function.settings = settings

        result = self.stash.upsert(context.credentials, endpoint=api_endpoint)
        if result.is_err():
            return SyftError(message=result.err())
        return SyftSuccess(message=f"APIEndpoint {api_path} settings updated.")

    @service_method(
        path="api.api_endpoints",
        name="api_endpoints",
        roles=DATA_SCIENTIST_ROLE_LEVEL,
    )
    def api_endpoints(
        self,
        context: AuthedServiceContext,
    ) -> list[TwinAPIEndpointView] | SyftError:
        """Retrieves a list of available API endpoints view available to the user."""
        admin_key = context.server.get_service("userservice").admin_verify_key()
        result = self.stash.get_all(admin_key)
        if result.is_err():
            return SyftError(message=result.err())

        all_api_endpoints = result.ok()
        api_endpoint_view = [
            api_endpoint.to(TwinAPIEndpointView, context=context)
            for api_endpoint in all_api_endpoints
        ]

        return api_endpoint_view

    @service_method(
        path="api.call_in_jobs", name="call_in_jobs", roles=GUEST_ROLE_LEVEL
    )
    def call_in_jobs(
        self,
        context: AuthedServiceContext,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> Any | SyftError:
        """Call a Custom API Method in a Job"""
        return self._call_in_jobs(context, "call", path, *args, **kwargs)

    @service_method(
        path="api.call_private_in_jobs",
        name="call_private_in_jobs",
        roles=GUEST_ROLE_LEVEL,
    )
    def call_private_in_jobs(
        self,
        context: AuthedServiceContext,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> Any | SyftError:
        """Call a Custom API Method in a Job"""
        return self._call_in_jobs(context, "call_private", path, *args, **kwargs)

    @service_method(
        path="api.call_public_in_jobs",
        name="call_public_in_jobs",
        roles=GUEST_ROLE_LEVEL,
    )
    def call_public_in_jobs(
        self,
        context: AuthedServiceContext,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> Any | SyftError:
        """Call a Custom API Method in a Job"""
        return self._call_in_jobs(context, "call_public", path, *args, **kwargs)

    def _call_in_jobs(
        self,
        context: AuthedServiceContext,
        method: str,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> Any | SyftError:
        custom_endpoint = self.get_code(
            context=context,
            endpoint_path=path,
        )
        if isinstance(custom_endpoint, SyftError):
            return custom_endpoint

        result = context.server.add_api_endpoint_execution_to_queue(
            context.credentials,
            method,
            path,
            *args,
            worker_pool=custom_endpoint.worker_pool,
            **kwargs,
        )
        if isinstance(result, SyftError):
            return result
        # relative
        from ..job.job_stash import JobStatus

        # So result is a Job object
        job = result
        job_service = context.server.get_service("jobservice")
        job_id = job.id
        # Question: For a small moment, when job status is updated, it doesn't return the job during the .get() as if
        # it's not in the stash. Then afterwards if appears again. Is this a bug?

        start = time.time()
        # TODO: what can we do here?????
        while (
            job is None
            or job.status == JobStatus.PROCESSING
            or job.status == JobStatus.CREATED
        ):
            job = job_service.get(context, job_id)
            time.sleep(0.1)
            if (time.time() - custom_endpoint.endpoint_timeout) > start:
                return SyftError(
                    message=(
                        f"Function timed out in {custom_endpoint.endpoint_timeout} seconds. "
                        + f"Get the Job with id: {job_id} to check results."
                    )
                )

        if job.status == JobStatus.COMPLETED:
            return job.result
        else:
            return SyftError(message="Function failed to complete.")

    @service_method(
        path="api.get_public_context", name="get_public_context", roles=ADMIN_ROLE_LEVEL
    )
    def get_public_context(
        self, context: AuthedServiceContext, path: str
    ) -> dict[str, Any] | SyftError:
        """Get specific public api context."""
        custom_endpoint = self.get_code(
            context=context,
            endpoint_path=path,
        )
        if isinstance(custom_endpoint, SyftError):
            return custom_endpoint

        return custom_endpoint.mock_function.build_internal_context(context=context).to(
            TwinAPIContextView
        )

    @service_method(
        path="api.get_private_context",
        name="get_private_context",
        roles=ADMIN_ROLE_LEVEL,
    )
    def get_private_context(
        self, context: AuthedServiceContext, path: str
    ) -> dict[str, Any] | SyftError:
        """Get specific private api context."""
        custom_endpoint = self.get_code(
            context=context,
            endpoint_path=path,
        )
        if isinstance(custom_endpoint, SyftError):
            return custom_endpoint

        custom_endpoint.private_function = cast(
            PrivateAPIEndpoint, custom_endpoint.private_function
        )

        return custom_endpoint.private_function.build_internal_context(
            context=context
        ).to(TwinAPIContextView)

    @service_method(path="api.get_all", name="get_all", roles=ADMIN_ROLE_LEVEL)
    def get_all(
        self,
        context: AuthedServiceContext,
    ) -> list[TwinAPIEndpoint] | SyftError:
        """Get all API endpoints."""
        result = self.stash.get_all(context.credentials)
        if result.is_ok():
            return result.ok()
        return SyftError(message=result.err())

    @service_method(path="api.call", name="call", roles=GUEST_ROLE_LEVEL)
    def call(
        self,
        context: AuthedServiceContext,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> SyftSuccess | SyftError:
        """Call a Custom API Method"""
        custom_endpoint = self.get_code(
            context=context,
            endpoint_path=path,
        )
        if isinstance(custom_endpoint, SyftError):
            return custom_endpoint

        exec_result = custom_endpoint.exec(context, *args, **kwargs)

        if isinstance(exec_result, SyftError):
            return Ok(exec_result)

        action_obj = ActionObject.from_obj(exec_result)
        action_service = cast(ActionService, context.server.get_service(ActionService))
        try:
            result = action_service.set_result_to_store(
                context=context,
                result_action_object=action_obj,
                has_result_read_permission=True,
            )
            if result.is_err():
                return SyftError(
                    message=f"Failed to set result to store: {result.err()}"
                )

            return Ok(result.ok())
        except Exception as e:
            # stdlib
            import traceback

            return Err(value=f"Failed to run. {e}, {traceback.format_exc()}")

    @service_method(path="api.call_public", name="call_public", roles=GUEST_ROLE_LEVEL)
    def call_public(
        self,
        context: AuthedServiceContext,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> ActionObject | SyftError:
        """Call a Custom API Method in public mode"""
        custom_endpoint = self.get_code(
            context=context,
            endpoint_path=path,
        )
        if isinstance(custom_endpoint, SyftError):
            return custom_endpoint
        exec_result = custom_endpoint.exec_mock_function(context, *args, **kwargs)

        if isinstance(exec_result, SyftError):
            return Ok(exec_result)

        action_obj = ActionObject.from_obj(exec_result)
        action_service = cast(ActionService, context.server.get_service(ActionService))
        try:
            result = action_service.set_result_to_store(
                context=context,
                result_action_object=action_obj,
                has_result_read_permission=True,
            )
            if result.is_err():
                return SyftError(
                    message=f"Failed to set result to store: {result.err()}"
                )

            return Ok(result.ok())
        except Exception as e:
            # stdlib
            import traceback

            return Err(value=f"Failed to run. {e}, {traceback.format_exc()}")

    @service_method(
        path="api.call_private", name="call_private", roles=GUEST_ROLE_LEVEL
    )
    def call_private(
        self,
        context: AuthedServiceContext,
        path: str,
        *args: Any,
        **kwargs: Any,
    ) -> ActionObject | SyftError:
        """Call a Custom API Method in private mode"""
        custom_endpoint = self.get_code(
            context=context,
            endpoint_path=path,
        )
        if isinstance(custom_endpoint, SyftError):
            return custom_endpoint

        exec_result = custom_endpoint.exec_private_function(context, *args, **kwargs)

        if isinstance(exec_result, SyftError):
            return Ok(exec_result)

        action_obj = ActionObject.from_obj(exec_result)

        action_service = cast(ActionService, context.server.get_service(ActionService))
        try:
            result = action_service.set_result_to_store(
                context=context, result_action_object=action_obj
            )
            if result.is_err():
                return SyftError(
                    message=f"Failed to set result to store: {result.err()}"
                )

            return Ok(result.ok())

        except Exception as e:
            # stdlib
            import traceback

            return Err(value=f"Failed to run. {e}, {traceback.format_exc()}")

    @service_method(
        path="api.exists",
        name="exists",
    )
    def exists(
        self, context: AuthedServiceContext, uid: UID
    ) -> SyftSuccess | SyftError:
        """Check if an endpoint exists"""
        endpoint = self.get_endpoint_by_uid(context, uid)
        return (
            SyftSuccess(message="Endpoint exists")
            if not isinstance(endpoint, SyftError)
            else endpoint
        )

    # ==== The methods below aren't meant to be called directly by the user, but
    # rather by the server context. ===
    # Therefore, they are not decorated with @service_method
    def execute_server_side_endpoint_by_id(
        self,
        context: AuthedServiceContext,
        endpoint_uid: UID,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        endpoint = self.get_endpoint_by_uid(context, endpoint_uid)
        if isinstance(endpoint, SyftError):
            return endpoint
        selected_code = endpoint.private_function
        if not selected_code:
            selected_code = endpoint.mock_function

        return endpoint.exec_code(selected_code, context, *args, **kwargs)

    def execute_service_side_endpoint_private_by_id(
        self,
        context: AuthedServiceContext,
        endpoint_uid: UID,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        endpoint = self.get_endpoint_by_uid(context, endpoint_uid)
        if isinstance(endpoint, SyftError):
            return endpoint
        if not endpoint.private_function:
            return SyftError(message="This endpoint does not have a private code")
        return endpoint.exec_code(endpoint.private_function, context, *args, **kwargs)

    def execute_server_side_endpoint_mock_by_id(
        self,
        context: AuthedServiceContext,
        endpoint_uid: UID,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        endpoint = self.get_endpoint_by_uid(context, endpoint_uid)
        if isinstance(endpoint, SyftError):
            return endpoint
        return endpoint.exec_code(endpoint.mock_function, context, *args, **kwargs)

    def get_endpoint_by_uid(
        self, context: AuthedServiceContext, uid: UID
    ) -> TwinAPIEndpoint | SyftError:
        admin_key = context.server.get_service("userservice").admin_verify_key()
        result = self.stash.get_by_uid(admin_key, uid)
        if result.is_err():
            return SyftError(message=result.err())
        return result.ok()

    def get_endpoints(
        self, context: AuthedServiceContext
    ) -> list[TwinAPIEndpoint] | SyftError:
        # TODO: Add ability to specify which roles see which endpoints
        # for now skip auth
        results = self.stash.get_all(context.server.verify_key)
        if results.is_ok():
            return results.ok()
        return SyftError(messages="Unable to get CustomAPIEndpoint")

    def get_code(
        self, context: AuthedServiceContext, endpoint_path: str
    ) -> TwinAPIEndpoint | SyftError:
        result = self.stash.get_by_path(context.server.verify_key, path=endpoint_path)
        if result.is_err():
            return SyftError(
                message=f"CustomAPIEndpoint: {endpoint_path} does not exist"
            )

        if result.is_ok():
            return result.ok()

        return SyftError(message=f"Unable to get {endpoint_path} CustomAPIEndpoint")


TYPE_TO_SERVICE[TwinAPIEndpoint] = APIService



# stdlib
from collections.abc import Callable

# relative
from ...types.transforms import TransformContext


def _downgrade_metadata_v3_to_v2() -> Callable:
    def set_defaults_from_settings(context: TransformContext) -> TransformContext:
        # Extract from settings if server is attached to context
        if context.output is not None:
            if context.server is not None:
                context.output["deployed_on"] = context.server.settings.deployed_on
                context.output["on_board"] = context.server.settings.on_board
                context.output["signup_enabled"] = (
                    context.server.settings.signup_enabled
                )
                context.output["admin_email"] = context.server.settings.admin_email
            else:
                # Else set default value
                context.output["signup_enabled"] = False
                context.output["admin_email"] = ""

        return context

    return set_defaults_from_settings



# stdlib
from importlib.resources import as_file
from importlib.resources import files

# third party
import capnp

# relative
from ..util._std_stream_capture import std_stream_capture


def get_capnp_schema(schema_file: str) -> type:
    with as_file(files("syft.capnp").joinpath(schema_file)) as capnp_path:
        with std_stream_capture():
            return capnp.load(str(capnp_path.absolute()))
